import base64
import json
import logging
from concurrent.futures import ThreadPoolExecutor, wait
from dataclasses import dataclass
from typing import Literal, Optional, Sequence, Union

import boto3
import pyarrow as pa
from botocore.config import Config
from pydantic import StrictStr

from feast import utils
from feast.batch_feature_view import BatchFeatureView
from feast.constants import FEATURE_STORE_YAML_ENV_NAME
from feast.entity import Entity
from feast.feature_view import FeatureView
from feast.infra.common.materialization_job import (
    MaterializationJob,
    MaterializationJobStatus,
    MaterializationTask,
)
from feast.infra.common.retrieval_task import HistoricalRetrievalTask
from feast.infra.compute_engines.base import ComputeEngine
from feast.infra.offline_stores.offline_store import OfflineStore
from feast.infra.online_stores.online_store import OnlineStore
from feast.infra.registry.base_registry import BaseRegistry
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.repo_config import FeastConfigBaseModel, RepoConfig
from feast.stream_feature_view import StreamFeatureView
from feast.utils import _get_column_names
from feast.version import get_version

DEFAULT_BATCH_SIZE = 10_000
DEFAULT_TIMEOUT = 600
LAMBDA_TIMEOUT_RETRIES = 5

logger = logging.getLogger(__name__)


class LambdaComputeEngineConfig(FeastConfigBaseModel):
    """Batch Compute Engine config for lambda based engine"""

    type: Literal["lambda"] = "lambda"
    """ Type selector"""

    materialization_image: StrictStr
    """ The URI of a container image in the Amazon ECR registry, which should be used for materialization. """

    lambda_role: StrictStr
    """ Role that should be used by the materialization lambda """


@dataclass
class LambdaMaterializationJob(MaterializationJob):
    def __init__(
        self,
        job_id: str,
        status: MaterializationJobStatus,
        error: Optional[BaseException] = None,
    ) -> None:
        super().__init__()
        self._job_id: str = job_id
        self._status = status
        self._error = error

    def status(self) -> MaterializationJobStatus:
        return self._status

    def error(self) -> Optional[BaseException]:
        return self._error

    def should_be_retried(self) -> bool:
        return False

    def job_id(self) -> str:
        return self._job_id

    def url(self) -> Optional[str]:
        return None


class LambdaComputeEngine(ComputeEngine):
    """
    WARNING: This engine should be considered "Alpha" functionality.
    """

    def get_historical_features(
        self, registry: BaseRegistry, task: HistoricalRetrievalTask
    ) -> pa.Table:
        raise NotImplementedError(
            "Lambda Compute Engine does not support get_historical_features"
        )

    def update(
        self,
        project: str,
        views_to_delete: Sequence[
            Union[BatchFeatureView, StreamFeatureView, FeatureView, OnDemandFeatureView]
        ],
        views_to_keep: Sequence[
            Union[BatchFeatureView, StreamFeatureView, FeatureView, OnDemandFeatureView]
        ],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
    ):
        # This should be setting up the lambda function.
        r = self.lambda_client.create_function(
            FunctionName=self.lambda_name,
            PackageType="Image",
            Role=self.repo_config.batch_engine.lambda_role,
            Code={"ImageUri": self.repo_config.batch_engine.materialization_image},
            Timeout=DEFAULT_TIMEOUT,
            Tags={
                "feast-owned": "True",
                "project": project,
                "feast-sdk-version": get_version(),
            },
        )
        logger.info(
            "Creating lambda function %s, %s",
            self.lambda_name,
            r["ResponseMetadata"]["RequestId"],
        )

        logger.info("Waiting for function %s to be active", self.lambda_name)
        waiter = self.lambda_client.get_waiter("function_active")
        waiter.wait(FunctionName=self.lambda_name)

    def teardown_infra(
        self,
        project: str,
        fvs: Sequence[Union[BatchFeatureView, StreamFeatureView, FeatureView]],
        entities: Sequence[Entity],
    ):
        # This should be tearing down the lambda function.
        logger.info("Tearing down lambda %s", self.lambda_name)
        r = self.lambda_client.delete_function(FunctionName=self.lambda_name)
        logger.info("Finished tearing down lambda %s: %s", self.lambda_name, r)

    def __init__(
        self,
        *,
        repo_config: RepoConfig,
        offline_store: OfflineStore,
        online_store: OnlineStore,
        **kwargs,
    ):
        super().__init__(
            repo_config=repo_config,
            offline_store=offline_store,
            online_store=online_store,
            **kwargs,
        )
        repo_path = self.repo_config.repo_path
        assert repo_path
        feature_store_path = utils.get_default_yaml_file_path(repo_path)
        self.feature_store_base64 = str(
            base64.b64encode(bytes(feature_store_path.read_text(), "UTF-8")), "UTF-8"
        )

        self.lambda_name = f"feast-materialize-{self.repo_config.project}"
        if len(self.lambda_name) > 64:
            self.lambda_name = self.lambda_name[:64]
        config = Config(read_timeout=DEFAULT_TIMEOUT + 10)
        self.lambda_client = boto3.client("lambda", config=config)

    def _materialize_one(
        self, registry: BaseRegistry, task: MaterializationTask, **kwargs
    ):
        feature_view = task.feature_view
        start_date = task.start_time
        end_date = task.end_time
        project = task.project

        entities = []
        for entity_name in feature_view.entities:
            entities.append(registry.get_entity(entity_name, project))

        (
            join_key_columns,
            feature_name_columns,
            timestamp_field,
            created_timestamp_column,
        ) = _get_column_names(feature_view, entities)

        job_id = f"{feature_view.name}-{start_date}-{end_date}"

        offline_job = self.offline_store.pull_latest_from_table_or_query(
            config=self.repo_config,
            data_source=feature_view.batch_source,
            join_key_columns=join_key_columns,
            feature_name_columns=feature_name_columns,
            timestamp_field=timestamp_field,
            created_timestamp_column=created_timestamp_column,
            start_date=start_date,
            end_date=end_date,
        )

        paths = offline_job.to_remote_storage()
        if (num_files := len(paths)) == 0:
            logger.warning("No values to update for the given time range.")
            return LambdaMaterializationJob(
                job_id=job_id, status=MaterializationJobStatus.SUCCEEDED
            )
        else:
            max_workers = num_files if num_files <= 20 else 20
            executor = ThreadPoolExecutor(max_workers=max_workers)
            futures = []

            for path in paths:
                payload = {
                    FEATURE_STORE_YAML_ENV_NAME: self.feature_store_base64,
                    "view_name": feature_view.name,
                    "view_type": "batch",
                    "path": path,
                }
                # Invoke a lambda to materialize this file.

                logger.info("Invoking materialization for %s", path)
                futures.append(
                    executor.submit(
                        self.invoke_with_retries,
                        FunctionName=self.lambda_name,
                        InvocationType="RequestResponse",
                        Payload=json.dumps(payload),
                    )
                )

            done, not_done = wait(futures)
            logger.info("Done: %s Not Done: %s", done, not_done)
            errors = []
            for f in done:
                response, payload = f.result()

                logger.info(
                    f"Ingested task; request id {response['ResponseMetadata']['RequestId']}, "
                    f"Output: {payload}"
                )
                if "errorMessage" in payload.keys():
                    errors.append(payload["errorMessage"])

            for f in not_done:
                response, payload = f.result()
                logger.error(f"Ingestion failed: {response=}, {payload=}")

            if len(not_done) == 0 and len(errors) == 0:
                return LambdaMaterializationJob(
                    job_id=job_id, status=MaterializationJobStatus.SUCCEEDED
                )
            else:
                return LambdaMaterializationJob(
                    job_id=job_id,
                    status=MaterializationJobStatus.ERROR,
                    error=RuntimeError(
                        f"Lambda functions did not finish successfully: {errors}"
                    ),
                )

    def invoke_with_retries(self, **kwargs):
        """Invoke the Lambda function and retry if it times out.

        The Lambda function may time out initially if many values are updated
        and DynamoDB throttles requests. As soon as the DynamoDB tables
        are scaled up, the Lambda function can succeed upon retry with higher
        throughput.

        """
        retries = 0
        while retries < LAMBDA_TIMEOUT_RETRIES:
            response = self.lambda_client.invoke(**kwargs)
            payload = json.loads(response["Payload"].read()) or {}
            if "Task timed out after" not in payload.get("errorMessage", ""):
                break
            retries += 1
            logger.warning(
                "Retrying lambda function after lambda timeout in request"
                f"{response['ResponseMetadata']['RequestId']}"
            )
        return response, payload



# Copyright 2024 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Image processing utilities for Feast image search capabilities.
Provides image embedding generation and combination functions for multi-modal search.
"""

import io
from typing import List

try:
    import timm
    import torch
    from PIL import Image
    from sklearn.preprocessing import normalize
    from timm.data import resolve_data_config
    from timm.data.transforms_factory import create_transform

    _image_dependencies_available = True
except ImportError:
    _image_dependencies_available = False


COMBINATION_STRATEGIES = ["weighted_sum", "concatenate", "average"]


def _check_image_dependencies():
    """Check if image processing dependencies are available."""
    if not _image_dependencies_available:
        raise ImportError(
            "Image processing dependencies are not installed. "
            "Please install with: pip install feast[image]"
        )


class ImageFeatureExtractor:
    """
    Extract image embeddings using pre-trained vision models.
    This class uses timm (PyTorch Image Models) to generate embeddings
    from images using pre-trained vision models like ResNet, ViT, etc.

    Examples:
        Basic usage::

            extractor = ImageFeatureExtractor()
            with open("image.jpg", "rb") as f:
                image_bytes = f.read()
            embedding = extractor.extract_embedding(image_bytes)

        Using different models::

            # ResNet-50
            extractor = ImageFeatureExtractor("resnet50")
            embedding = extractor.extract_embedding(image_bytes)
            # ViT model
            extractor = ImageFeatureExtractor("vit_base_patch16_224")
            embedding = extractor.extract_embedding(image_bytes)
    """

    def __init__(self, model_name: str = "resnet34"):
        """
        Initialize with a pre-trained model.
        Args:
            model_name: Model name from timm library. Popular choices:
                - "resnet34": Fast, good for general use (default)
                - "resnet50": Better accuracy than ResNet-34
                - "vit_base_patch16_224": Vision Transformer, high accuracy
                - "efficientnet_b0": Good balance of speed and accuracy
                - "mobilenetv3_large_100": Fast inference for mobile/edge
        Raises:
            ImportError: If image processing dependencies are not installed
            RuntimeError: If the specified model cannot be loaded
        """
        _check_image_dependencies()

        try:
            self.model_name = model_name
            self.model = timm.create_model(
                model_name, pretrained=True, num_classes=0, global_pool="avg"
            )
            self.model.eval()

            config = resolve_data_config({}, model=model_name)
            self.preprocess = create_transform(**config)

        except Exception as e:
            raise RuntimeError(f"Failed to load model '{model_name}': {e}")

    def extract_embedding(self, image_bytes: bytes) -> List[float]:
        """
        Extract embedding from image bytes.
        Args:
            image_bytes: Image data as bytes (JPEG, PNG, WebP, etc.)
        Returns:
            Normalized embedding vector as list of floats
        Raises:
            ValueError: If image cannot be processed or is invalid
        """
        try:
            image = Image.open(io.BytesIO(image_bytes)).convert("RGB")
            input_tensor = self.preprocess(image).unsqueeze(0)

            with torch.no_grad():
                output = self.model(input_tensor)

            feature_vector = output.squeeze().numpy()
            normalized = normalize(feature_vector.reshape(1, -1), norm="l2")
            return normalized.flatten().tolist()

        except Exception as e:
            raise ValueError(f"Failed to extract embedding from image: {e}")

    def batch_extract_embeddings(
        self, image_bytes_list: List[bytes]
    ) -> List[List[float]]:
        """
        Extract embeddings from multiple images in batch for efficiency.
        Args:
            image_bytes_list: List of image data as bytes
        Returns:
            List of normalized embedding vectors
        Raises:
            ValueError: If any image cannot be processed
        """
        embeddings = []

        images = []
        for image_bytes in image_bytes_list:
            try:
                image = Image.open(io.BytesIO(image_bytes)).convert("RGB")
                preprocessed = self.preprocess(image)
                images.append(preprocessed)
            except Exception as e:
                raise ValueError(f"Failed to preprocess image: {e}")

        batch_tensor = torch.stack(images)

        with torch.no_grad():
            outputs = self.model(batch_tensor)

        for output in outputs:
            feature_vector = output.numpy()
            normalized = normalize(feature_vector.reshape(1, -1), norm="l2")
            embeddings.append(normalized.flatten().tolist())

        return embeddings


def combine_embeddings(
    text_embedding: List[float],
    image_embedding: List[float],
    strategy: str = "weighted_sum",
    text_weight: float = 0.5,
    image_weight: float = 0.5,
) -> List[float]:
    """
    Combine text and image embeddings search.
    This function provides several strategies for combining embeddings from
    different modalities (text and image) into a single vector for search.

    Args:
        text_embedding: Text embedding vector
        image_embedding: Image embedding vector
        strategy: Combination strategy (default: "weighted_sum")
        text_weight: Weight for text embedding (for weighted strategies)
        image_weight: Weight for image embedding (for weighted strategies)

    Returns:
        Combined embedding vector as list of floats

    Raises:
        ValueError: If weights don't sum to 1.0 for weighted_sum strategy

    Examples:
        Weighted combination (emphasize image)::

            combined = combine_embeddings(
                [0.1, 0.2], [0.8, 0.9],  # text_emb, image_emb
                strategy="weighted_sum",
                text_weight=0.3, image_weight=0.7
            )

        Concatenation for full information::

            combined = combine_embeddings(
                [0.1, 0.2], [0.8, 0.9],  # text_emb, image_emb
                strategy="concatenate"
            )
    """
    if strategy == "weighted_sum":
        if abs(text_weight + image_weight - 1.0) > 1e-6:
            raise ValueError(
                "text_weight + image_weight must equal 1.0 for weighted_sum"
            )

        max_dim = max(len(text_embedding), len(image_embedding))
        text_padded = text_embedding + [0.0] * (max_dim - len(text_embedding))
        image_padded = image_embedding + [0.0] * (max_dim - len(image_embedding))

        combined = [
            text_weight * t + image_weight * i
            for t, i in zip(text_padded, image_padded)
        ]
        return combined

    elif strategy == "concatenate":
        return text_embedding + image_embedding

    elif strategy == "average":
        max_dim = max(len(text_embedding), len(image_embedding))
        text_padded = text_embedding + [0.0] * (max_dim - len(text_embedding))
        image_padded = image_embedding + [0.0] * (max_dim - len(image_embedding))

        combined = [(t + i) / 2.0 for t, i in zip(text_padded, image_padded)]
        return combined

    else:
        raise ValueError(
            f"Unknown combination strategy: {strategy}. "
            f"Supported strategies: {', '.join(COMBINATION_STRATEGIES)}"
        )


def validate_image_format(image_bytes: bytes) -> bool:
    """
    Validate that the provided bytes represent a valid image.
    Args:
        image_bytes: Image data as bytes
    Returns:
        True if valid image, False otherwise
    """
    try:
        with Image.open(io.BytesIO(image_bytes)) as img:
            img.verify()
        return True
    except Exception:
        return False


def get_image_metadata(image_bytes: bytes) -> dict:
    """
    Extract metadata from image bytes.
    Args:
        image_bytes: Image data as bytes
    Returns:
        Dictionary with image metadata (format, size, mode, etc.)
    Raises:
        ValueError: If image cannot be processed
    """
    try:
        with Image.open(io.BytesIO(image_bytes)) as img:
            return {
                "format": img.format,
                "mode": img.mode,
                "width": img.width,
                "height": img.height,
                "size_bytes": len(image_bytes),
            }
    except Exception as e:
        raise ValueError(f"Failed to extract image metadata: {e}")



import logging
from typing import Optional

logger = logging.getLogger(__name__)


class User:
    _username: str
    _roles: Optional[list[str]]
    _groups: Optional[list[str]]
    _namespaces: Optional[list[str]]

    def __init__(
        self,
        username: str,
        roles: list[str] = [],
        groups: list[str] = [],
        namespaces: list[str] = [],
    ):
        self._username = username
        self._roles = roles
        self._groups = groups
        self._namespaces = namespaces

    @property
    def username(self):
        return self._username

    @property
    def roles(self):
        return self._roles

    @property
    def groups(self):
        return self._groups

    @property
    def namespaces(self):
        return self._namespaces

    def has_matching_role(self, requested_roles: list[str]) -> bool:
        """
        Verify that the user has at least one of the requested roles.

        Args:
            requested_roles: The list of requested roles.

        Returns:
            bool: `True` only if the user has any registered role and all the given roles are registered.
        """
        logger.debug(
            f"Check {self.username} has all {requested_roles}: currently {self.roles}"
        )
        return any(role in self.roles for role in requested_roles)

    def has_matching_group(self, requested_groups: list[str]) -> bool:
        """
        Verify that the user has at least one of the requested groups.

        Args:
            requested_groups: The list of requested groups.

        Returns:
            bool: `True` only if the user has any registered group and all the given groups are registered.
        """
        logger.debug(
            f"Check {self.username} has all {requested_groups}: currently {self.groups}"
        )
        return any(group in self.groups for group in requested_groups)

    def has_matching_namespace(self, requested_namespaces: list[str]) -> bool:
        """
        Verify that the user has at least one of the requested namespaces.

        Args:
            requested_namespaces: The list of requested namespaces.

        Returns:
            bool: `True` only if the user has any registered namespace and all the given namespaces are registered.
        """
        logger.debug(
            f"Check {self.username} has all {requested_namespaces}: currently {self.namespaces}"
        )
        return any(namespace in self.namespaces for namespace in requested_namespaces)

    def __str__(self):
        return f"{self.username} (roles: {self.roles}, groups: {self.groups}, namespaces: {self.namespaces})"



import logging

import jwt

from feast.permissions.auth.auth_type import AuthType
from feast.permissions.auth_model import AuthConfig
from feast.permissions.client.auth_client_manager import AuthenticationClientManager

logger = logging.getLogger(__name__)


class IntraCommAuthClientManager(AuthenticationClientManager):
    def __init__(self, auth_config: AuthConfig, intra_communication_base64: str):
        self.auth_config = auth_config
        self.intra_communication_base64 = intra_communication_base64
        logger.debug(f"AuthConfig type set to {self.auth_config.type}")

    def get_token(self):
        if self.auth_config.type == AuthType.OIDC.value:
            payload = {
                "preferred_username": f"{self.intra_communication_base64}",  # Subject claim
            }
        elif self.auth_config.type == AuthType.KUBERNETES.value:
            payload = {
                "sub": f":::{self.intra_communication_base64}",  # Subject claim
            }
        else:
            raise RuntimeError(
                f"No Auth client manager implemented for the auth type:{self.auth_config.type}"
            )

        return jwt.encode(payload, "")



import json
import uuid
from pathlib import Path
from typing import Optional

from feast.infra.registry.registry_store import RegistryStore
from feast.protos.feast.core.Registry_pb2 import Registry as RegistryProto
from feast.repo_config import RegistryConfig
from feast.utils import _utc_now


class FileRegistryStore(RegistryStore):
    def __init__(self, registry_config: RegistryConfig, repo_path: Path):
        registry_path = Path(registry_config.path)
        if registry_path.is_absolute():
            self._filepath = registry_path
        else:
            self._filepath = repo_path.joinpath(registry_path)

    def get_registry_proto(self):
        registry_proto = RegistryProto()
        if self._filepath.exists():
            registry_proto.ParseFromString(self._filepath.read_bytes())
            return registry_proto
        raise FileNotFoundError(
            f'Registry not found at path "{self._filepath}". Have you run "feast apply"?'
        )

    def update_registry_proto(self, registry_proto: RegistryProto):
        self._write_registry(registry_proto)

    def teardown(self):
        try:
            self._filepath.unlink()
        except FileNotFoundError:
            # If the file deletion fails with FileNotFoundError, the file has already
            # been deleted.
            pass

    def _write_registry(self, registry_proto: RegistryProto):
        registry_proto.version_id = str(uuid.uuid4())
        registry_proto.last_updated.FromDatetime(_utc_now())
        file_dir = self._filepath.parent
        file_dir.mkdir(exist_ok=True)
        with open(self._filepath, mode="wb", buffering=0) as f:
            f.write(registry_proto.SerializeToString())

    def set_project_metadata(self, project: str, key: str, value: str):
        """Set a custom project metadata key-value pair in the registry proto (file backend)."""
        registry_proto = self.get_registry_proto()
        found = False
        for pm in registry_proto.project_metadata:
            if pm.project == project:
                # Use a special key for custom metadata
                if hasattr(pm, "custom_metadata"):
                    # If custom_metadata is a map<string, string>
                    pm.custom_metadata[key] = value
                else:
                    # Fallback: store as JSON in a special key
                    try:
                        meta = json.loads(pm.project_uuid) if pm.project_uuid else {}
                    except Exception:
                        meta = {}
                    if not isinstance(meta, dict):
                        meta = {}
                    meta[key] = value
                    pm.project_uuid = json.dumps(meta)
                found = True
                break
        if not found:
            from feast.project_metadata import ProjectMetadata

            pm = ProjectMetadata(project_name=project)
            # Fallback: store as JSON in project_uuid
            pm.project_uuid = json.dumps({key: value})
            registry_proto.project_metadata.append(pm.to_proto())
        self.update_registry_proto(registry_proto)

    def get_project_metadata(self, project: str, key: str) -> Optional[str]:
        """Get a custom project metadata value by key from the registry proto (file backend)."""
        registry_proto = self.get_registry_proto()
        for pm in registry_proto.project_metadata:
            if pm.project == project:
                if hasattr(pm, "custom_metadata"):
                    return pm.custom_metadata.get(key, None)
                else:
                    try:
                        meta = json.loads(pm.project_uuid) if pm.project_uuid else {}
                    except Exception:
                        meta = {}
                    if not isinstance(meta, dict):
                        return None
                    return meta.get(key, None)
        return None



import logging
from concurrent import futures
from datetime import datetime, timezone
from typing import Any, List, Optional, Union, cast

import grpc
from google.protobuf.empty_pb2 import Empty
from grpc_health.v1 import health, health_pb2, health_pb2_grpc
from grpc_reflection.v1alpha import reflection

from feast import FeatureService, FeatureStore
from feast.base_feature_view import BaseFeatureView
from feast.data_source import DataSource
from feast.entity import Entity
from feast.errors import FeastObjectNotFoundException, FeatureViewNotFoundException
from feast.feast_object import FeastObject
from feast.feature_view import FeatureView
from feast.grpc_error_interceptor import ErrorInterceptor
from feast.infra.infra_object import Infra
from feast.infra.registry.base_registry import BaseRegistry
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.permissions.action import AuthzedAction
from feast.permissions.permission import Permission
from feast.permissions.security_manager import (
    assert_permissions,
    assert_permissions_to_update,
    permitted_resources,
)
from feast.permissions.server.grpc import AuthInterceptor
from feast.permissions.server.utils import (
    AuthManagerType,
    ServerType,
    init_auth_manager,
    init_security_manager,
    str_to_auth_manager_type,
)
from feast.project import Project
from feast.protos.feast.registry import RegistryServer_pb2, RegistryServer_pb2_grpc
from feast.protos.feast.registry.RegistryServer_pb2 import Feature, ListFeaturesResponse
from feast.saved_dataset import SavedDataset, ValidationReference
from feast.stream_feature_view import StreamFeatureView

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def apply_pagination_and_sorting(
    objects: List[Any],
    pagination: Optional[RegistryServer_pb2.PaginationParams] = None,
    sorting: Optional[RegistryServer_pb2.SortingParams] = None,
) -> tuple[List[Any], RegistryServer_pb2.PaginationMetadata]:
    """
    Apply sorting and pagination to a list of objects at the gRPC layer.

    Args:
        objects: List of objects to paginate and sort
        pagination: Pagination parameters (page, limit)
        sorting: Sorting parameters (sort_by, sort_order)

    Returns:
        Tuple of (paginated and sorted objects, pagination metadata)
    """
    if not objects:
        empty_metadata = RegistryServer_pb2.PaginationMetadata(
            page=0,
            limit=0,
            total_count=0,
            total_pages=0,
            has_next=False,
            has_previous=False,
        )
        return objects, empty_metadata

    if sorting and sorting.sort_by and sorting.sort_by.strip():
        objects = apply_sorting(objects, sorting.sort_by, sorting.sort_order)

    total_count = len(objects)

    # Apply pagination if requested
    if pagination and pagination.page > 0 and pagination.limit > 0:
        start_idx = (pagination.page - 1) * pagination.limit
        end_idx = start_idx + pagination.limit
        paginated_objects = objects[start_idx:end_idx]

        total_pages = (total_count + pagination.limit - 1) // pagination.limit
        has_next = pagination.page < total_pages
        has_previous = pagination.page > 1

        pagination_metadata = RegistryServer_pb2.PaginationMetadata(
            page=pagination.page,
            limit=pagination.limit,
            total_count=total_count,
            total_pages=total_pages,
            has_next=has_next,
            has_previous=has_previous,
        )
        return paginated_objects, pagination_metadata
    else:
        # No pagination requested - return all objects
        pagination_metadata = RegistryServer_pb2.PaginationMetadata(
            page=0,
            limit=0,
            total_count=total_count,
            total_pages=1,
            has_next=False,
            has_previous=False,
        )
        return objects, pagination_metadata


def apply_sorting(objects: List[Any], sort_by: str, sort_order: str) -> List[Any]:
    """Apply sorting to a list of objects using dot notation for nested attributes."""

    def get_nested_attr(obj, attr_path: str):
        """Get nested attribute using dot notation."""
        attrs = attr_path.split(".")
        current = obj
        for attr in attrs:
            if hasattr(current, attr):
                current = getattr(current, attr)
            elif isinstance(current, dict) and attr in current:
                current = current[attr]
            else:
                return None
        return current

    def sort_key(obj):
        value = get_nested_attr(obj, sort_by)
        if value is None:
            return ("", 0, None)

        if isinstance(value, str):
            return (value.lower(), 1, None)
        elif isinstance(value, (int, float)):
            return ("", 1, value)
        else:
            return (str(value).lower(), 1, None)

    reverse = sort_order.lower() == "desc"

    try:
        return sorted(objects, key=sort_key, reverse=reverse)
    except Exception as e:
        logger.warning(f"Failed to sort objects by '{sort_by}': {e}")
        return objects


def _build_any_feature_view_proto(feature_view: BaseFeatureView):
    if isinstance(feature_view, StreamFeatureView):
        arg_name = "stream_feature_view"
        feature_view_proto = feature_view.to_proto()
    elif isinstance(feature_view, FeatureView):
        arg_name = "feature_view"
        feature_view_proto = feature_view.to_proto()
    elif isinstance(feature_view, OnDemandFeatureView):
        arg_name = "on_demand_feature_view"
        feature_view_proto = feature_view.to_proto()

    return RegistryServer_pb2.AnyFeatureView(
        feature_view=feature_view_proto if arg_name == "feature_view" else None,
        stream_feature_view=feature_view_proto
        if arg_name == "stream_feature_view"
        else None,
        on_demand_feature_view=feature_view_proto
        if arg_name == "on_demand_feature_view"
        else None,
    )


class RegistryServer(RegistryServer_pb2_grpc.RegistryServerServicer):
    def __init__(self, registry: BaseRegistry) -> None:
        super().__init__()
        self.proxied_registry = registry

    def ApplyEntity(self, request: RegistryServer_pb2.ApplyEntityRequest, context):
        entity = cast(
            Entity,
            assert_permissions_to_update(
                resource=Entity.from_proto(request.entity),
                getter=self.proxied_registry.get_entity,
                project=request.project,
            ),
        )
        self.proxied_registry.apply_entity(
            entity=entity,
            project=request.project,
            commit=request.commit,
        )

        return Empty()

    def GetEntity(self, request: RegistryServer_pb2.GetEntityRequest, context):
        return assert_permissions(
            self.proxied_registry.get_entity(
                name=request.name,
                project=request.project,
                allow_cache=request.allow_cache,
            ),
            actions=[AuthzedAction.DESCRIBE],
        ).to_proto()

    def ListEntities(self, request: RegistryServer_pb2.ListEntitiesRequest, context):
        paginated_entities, pagination_metadata = apply_pagination_and_sorting(
            permitted_resources(
                resources=cast(
                    list[FeastObject],
                    self.proxied_registry.list_entities(
                        project=request.project,
                        allow_cache=request.allow_cache,
                        tags=dict(request.tags),
                    ),
                ),
                actions=AuthzedAction.DESCRIBE,
            ),
            pagination=request.pagination,
            sorting=request.sorting,
        )

        return RegistryServer_pb2.ListEntitiesResponse(
            entities=[entity.to_proto() for entity in paginated_entities],
            pagination=pagination_metadata,
        )

    def DeleteEntity(self, request: RegistryServer_pb2.DeleteEntityRequest, context):
        assert_permissions(
            resource=self.proxied_registry.get_entity(
                name=request.name, project=request.project
            ),
            actions=AuthzedAction.DELETE,
        )

        self.proxied_registry.delete_entity(
            name=request.name, project=request.project, commit=request.commit
        )
        return Empty()

    def ApplyDataSource(
        self, request: RegistryServer_pb2.ApplyDataSourceRequest, context
    ):
        data_source = cast(
            DataSource,
            assert_permissions_to_update(
                resource=DataSource.from_proto(request.data_source),
                getter=self.proxied_registry.get_data_source,
                project=request.project,
            ),
        )
        self.proxied_registry.apply_data_source(
            data_source=data_source,
            project=request.project,
            commit=request.commit,
        )

        return Empty()

    def GetDataSource(self, request: RegistryServer_pb2.GetDataSourceRequest, context):
        return assert_permissions(
            resource=self.proxied_registry.get_data_source(
                name=request.name,
                project=request.project,
                allow_cache=request.allow_cache,
            ),
            actions=AuthzedAction.DESCRIBE,
        ).to_proto()

    def ListDataSources(
        self, request: RegistryServer_pb2.ListDataSourcesRequest, context
    ):
        paginated_data_sources, pagination_metadata = apply_pagination_and_sorting(
            permitted_resources(
                resources=cast(
                    list[FeastObject],
                    self.proxied_registry.list_data_sources(
                        project=request.project,
                        allow_cache=request.allow_cache,
                        tags=dict(request.tags),
                    ),
                ),
                actions=AuthzedAction.DESCRIBE,
            ),
            pagination=request.pagination,
            sorting=request.sorting,
        )

        return RegistryServer_pb2.ListDataSourcesResponse(
            data_sources=[
                data_source.to_proto() for data_source in paginated_data_sources
            ],
            pagination=pagination_metadata,
        )

    def DeleteDataSource(
        self, request: RegistryServer_pb2.DeleteDataSourceRequest, context
    ):
        assert_permissions(
            resource=self.proxied_registry.get_data_source(
                name=request.name,
                project=request.project,
            ),
            actions=AuthzedAction.DELETE,
        )

        self.proxied_registry.delete_data_source(
            name=request.name, project=request.project, commit=request.commit
        )
        return Empty()

    def GetFeatureView(
        self, request: RegistryServer_pb2.GetFeatureViewRequest, context
    ):
        return assert_permissions(
            self.proxied_registry.get_feature_view(
                name=request.name,
                project=request.project,
                allow_cache=request.allow_cache,
            ),
            actions=[AuthzedAction.DESCRIBE],
        ).to_proto()

    def GetAnyFeatureView(
        self, request: RegistryServer_pb2.GetAnyFeatureViewRequest, context
    ):
        feature_view = assert_permissions(
            cast(
                FeastObject,
                self.proxied_registry.get_any_feature_view(
                    name=request.name,
                    project=request.project,
                    allow_cache=request.allow_cache,
                ),
            ),
            actions=[AuthzedAction.DESCRIBE],
        )

        return RegistryServer_pb2.GetAnyFeatureViewResponse(
            any_feature_view=_build_any_feature_view_proto(
                cast(BaseFeatureView, feature_view)
            )
        )

    def ApplyFeatureView(
        self, request: RegistryServer_pb2.ApplyFeatureViewRequest, context
    ):
        feature_view_type = request.WhichOneof("base_feature_view")
        if feature_view_type == "feature_view":
            feature_view = FeatureView.from_proto(request.feature_view)
        elif feature_view_type == "on_demand_feature_view":
            feature_view = OnDemandFeatureView.from_proto(
                request.on_demand_feature_view
            )
        elif feature_view_type == "stream_feature_view":
            feature_view = StreamFeatureView.from_proto(request.stream_feature_view)

        assert_permissions_to_update(
            resource=feature_view,
            # Will replace with the new get_any_feature_view method later
            getter=self.proxied_registry.get_feature_view,
            project=request.project,
        )

        (
            self.proxied_registry.apply_feature_view(
                feature_view=feature_view,
                project=request.project,
                commit=request.commit,
            ),
        )

        return Empty()

    def ListFeatureViews(
        self, request: RegistryServer_pb2.ListFeatureViewsRequest, context
    ):
        paginated_feature_views, pagination_metadata = apply_pagination_and_sorting(
            permitted_resources(
                resources=cast(
                    list[FeastObject],
                    self.proxied_registry.list_feature_views(
                        project=request.project,
                        allow_cache=request.allow_cache,
                        tags=dict(request.tags),
                    ),
                ),
                actions=AuthzedAction.DESCRIBE,
            ),
            pagination=request.pagination,
            sorting=request.sorting,
        )

        return RegistryServer_pb2.ListFeatureViewsResponse(
            feature_views=[
                feature_view.to_proto() for feature_view in paginated_feature_views
            ],
            pagination=pagination_metadata,
        )

    def ListAllFeatureViews(
        self, request: RegistryServer_pb2.ListAllFeatureViewsRequest, context
    ):
        all_feature_views = cast(
            list[FeastObject],
            self.proxied_registry.list_all_feature_views(
                project=request.project,
                allow_cache=request.allow_cache,
                tags=dict(request.tags),
            ),
        )

        if (
            request.entity
            or request.feature
            or request.feature_service
            or request.data_source
        ):
            filtered_feature_views = []
            for feature_view in all_feature_views:
                feature_view_obj = cast(BaseFeatureView, feature_view)
                include_feature_view = True

                # Filter by entity
                if request.entity:
                    entity_match = False
                    if (
                        hasattr(feature_view_obj, "entities")
                        and feature_view_obj.entities
                    ):
                        if request.entity in feature_view_obj.entities:
                            entity_match = True
                    elif (
                        hasattr(feature_view_obj, "entity_columns")
                        and feature_view_obj.entity_columns
                    ):
                        # For feature views with entity_columns, check if the entity is referenced
                        entity_names = [
                            col.entity_name
                            for col in feature_view_obj.entity_columns
                            if hasattr(col, "entity_name")
                        ]
                        if request.entity in entity_names:
                            entity_match = True
                    if not entity_match:
                        include_feature_view = False

                # Filter by feature
                if include_feature_view and request.feature:
                    feature_match = False
                    if hasattr(feature_view_obj, "features"):
                        feature_names = [f.name for f in feature_view_obj.features]
                        if request.feature in feature_names:
                            feature_match = True
                    if not feature_match:
                        include_feature_view = False

                # Filter by feature service
                if include_feature_view and request.feature_service:
                    try:
                        feature_service = self.proxied_registry.get_feature_service(
                            name=request.feature_service,
                            project=request.project,
                            allow_cache=request.allow_cache,
                        )
                        assert_permissions(
                            resource=feature_service,
                            actions=AuthzedAction.DESCRIBE,
                        )

                        feature_service_match = False
                        if hasattr(feature_service, "feature_view_projections"):
                            for (
                                feature_view_ref
                            ) in feature_service.feature_view_projections:
                                if feature_view_ref.name == feature_view_obj.name:
                                    feature_service_match = True
                                    break

                        if not feature_service_match:
                            include_feature_view = False
                    except Exception:
                        include_feature_view = False

                # Filter by data source
                if include_feature_view and request.data_source:
                    data_source_match = False
                    if (
                        hasattr(feature_view_obj, "batch_source")
                        and feature_view_obj.batch_source
                    ):
                        if feature_view_obj.batch_source.name == request.data_source:
                            data_source_match = True
                    if (
                        hasattr(feature_view_obj, "stream_source")
                        and feature_view_obj.stream_source
                    ):
                        if feature_view_obj.stream_source.name == request.data_source:
                            data_source_match = True
                    if (
                        hasattr(feature_view_obj, "source_request_sources")
                        and feature_view_obj.source_request_sources
                    ):
                        for (
                            request_source
                        ) in feature_view_obj.source_request_sources.values():
                            if request_source.name == request.data_source:
                                data_source_match = True
                                break
                    if not data_source_match:
                        include_feature_view = False

                if include_feature_view:
                    filtered_feature_views.append(feature_view_obj)

            all_feature_views = cast(list[FeastObject], filtered_feature_views)

        paginated_feature_views, pagination_metadata = apply_pagination_and_sorting(
            permitted_resources(
                resources=all_feature_views,
                actions=AuthzedAction.DESCRIBE,
            ),
            pagination=request.pagination,
            sorting=request.sorting,
        )

        return RegistryServer_pb2.ListAllFeatureViewsResponse(
            feature_views=[
                _build_any_feature_view_proto(cast(BaseFeatureView, feature_view))
                for feature_view in paginated_feature_views
            ],
            pagination=pagination_metadata,
        )

    def DeleteFeatureView(
        self, request: RegistryServer_pb2.DeleteFeatureViewRequest, context
    ):
        feature_view: Union[StreamFeatureView, FeatureView]

        try:
            feature_view = self.proxied_registry.get_stream_feature_view(
                name=request.name, project=request.project, allow_cache=False
            )
        except FeatureViewNotFoundException:
            feature_view = self.proxied_registry.get_feature_view(
                name=request.name, project=request.project, allow_cache=False
            )

        assert_permissions(
            resource=feature_view,
            actions=[AuthzedAction.DELETE],
        )
        self.proxied_registry.delete_feature_view(
            name=request.name, project=request.project, commit=request.commit
        )
        return Empty()

    def GetStreamFeatureView(
        self, request: RegistryServer_pb2.GetStreamFeatureViewRequest, context
    ):
        return assert_permissions(
            resource=self.proxied_registry.get_stream_feature_view(
                name=request.name,
                project=request.project,
                allow_cache=request.allow_cache,
            ),
            actions=[AuthzedAction.DESCRIBE],
        ).to_proto()

    def ListStreamFeatureViews(
        self, request: RegistryServer_pb2.ListStreamFeatureViewsRequest, context
    ):
        paginated_stream_feature_views, pagination_metadata = (
            apply_pagination_and_sorting(
                permitted_resources(
                    resources=cast(
                        list[FeastObject],
                        self.proxied_registry.list_stream_feature_views(
                            project=request.project,
                            allow_cache=request.allow_cache,
                            tags=dict(request.tags),
                        ),
                    ),
                    actions=AuthzedAction.DESCRIBE,
                ),
                pagination=request.pagination,
                sorting=request.sorting,
            )
        )

        return RegistryServer_pb2.ListStreamFeatureViewsResponse(
            stream_feature_views=[
                stream_feature_view.to_proto()
                for stream_feature_view in paginated_stream_feature_views
            ],
            pagination=pagination_metadata,
        )

    def GetOnDemandFeatureView(
        self, request: RegistryServer_pb2.GetOnDemandFeatureViewRequest, context
    ):
        return assert_permissions(
            resource=self.proxied_registry.get_on_demand_feature_view(
                name=request.name,
                project=request.project,
                allow_cache=request.allow_cache,
            ),
            actions=[AuthzedAction.DESCRIBE],
        ).to_proto()

    def ListOnDemandFeatureViews(
        self, request: RegistryServer_pb2.ListOnDemandFeatureViewsRequest, context
    ):
        paginated_on_demand_feature_views, pagination_metadata = (
            apply_pagination_and_sorting(
                permitted_resources(
                    resources=cast(
                        list[FeastObject],
                        self.proxied_registry.list_on_demand_feature_views(
                            project=request.project,
                            allow_cache=request.allow_cache,
                            tags=dict(request.tags),
                        ),
                    ),
                    actions=AuthzedAction.DESCRIBE,
                ),
                pagination=request.pagination,
                sorting=request.sorting,
            )
        )

        return RegistryServer_pb2.ListOnDemandFeatureViewsResponse(
            on_demand_feature_views=[
                on_demand_feature_view.to_proto()
                for on_demand_feature_view in paginated_on_demand_feature_views
            ],
            pagination=pagination_metadata,
        )

    def ApplyFeatureService(
        self, request: RegistryServer_pb2.ApplyFeatureServiceRequest, context
    ):
        feature_service = cast(
            FeatureService,
            assert_permissions_to_update(
                resource=FeatureService.from_proto(request.feature_service),
                getter=self.proxied_registry.get_feature_service,
                project=request.project,
            ),
        )
        self.proxied_registry.apply_feature_service(
            feature_service=feature_service,
            project=request.project,
            commit=request.commit,
        )

        return Empty()

    def GetFeatureService(
        self, request: RegistryServer_pb2.GetFeatureServiceRequest, context
    ):
        return assert_permissions(
            resource=self.proxied_registry.get_feature_service(
                name=request.name,
                project=request.project,
                allow_cache=request.allow_cache,
            ),
            actions=[AuthzedAction.DESCRIBE],
        ).to_proto()

    def ListFeatureServices(
        self, request: RegistryServer_pb2.ListFeatureServicesRequest, context
    ):
        # Get all feature services first
        all_feature_services = self.proxied_registry.list_feature_services(
            project=request.project,
            allow_cache=request.allow_cache,
            tags=dict(request.tags),
        )

        # Filter by feature view if specified
        if request.feature_view:
            filtered_feature_services = []
            for feature_service in all_feature_services:
                # Check if the feature service uses the specified feature view
                feature_view_match = False
                if hasattr(feature_service, "feature_view_projections"):
                    for (
                        feature_view_projection
                    ) in feature_service.feature_view_projections:
                        if feature_view_projection.name == request.feature_view:
                            feature_view_match = True
                            break
                if feature_view_match:
                    filtered_feature_services.append(feature_service)
            all_feature_services = filtered_feature_services

        paginated_feature_services, pagination_metadata = apply_pagination_and_sorting(
            permitted_resources(
                resources=cast(list[FeastObject], all_feature_services),
                actions=AuthzedAction.DESCRIBE,
            ),
            pagination=request.pagination,
            sorting=request.sorting,
        )

        return RegistryServer_pb2.ListFeatureServicesResponse(
            feature_services=[
                feature_service.to_proto()
                for feature_service in paginated_feature_services
            ],
            pagination=pagination_metadata,
        )

    def DeleteFeatureService(
        self, request: RegistryServer_pb2.DeleteFeatureServiceRequest, context
    ):
        feature_service = self.proxied_registry.get_feature_service(
            name=request.name, project=request.project, allow_cache=False
        )
        assert_permissions(
            resource=feature_service,
            actions=[AuthzedAction.DELETE],
        )
        self.proxied_registry.delete_feature_service(
            name=request.name, project=request.project, commit=request.commit
        )
        return Empty()

    def ApplySavedDataset(
        self, request: RegistryServer_pb2.ApplySavedDatasetRequest, context
    ):
        saved_dataset = cast(
            SavedDataset,
            assert_permissions_to_update(
                resource=SavedDataset.from_proto(request.saved_dataset),
                getter=self.proxied_registry.get_saved_dataset,
                project=request.project,
            ),
        )
        self.proxied_registry.apply_saved_dataset(
            saved_dataset=saved_dataset,
            project=request.project,
            commit=request.commit,
        )

        return Empty()

    def GetSavedDataset(
        self, request: RegistryServer_pb2.GetSavedDatasetRequest, context
    ):
        return assert_permissions(
            self.proxied_registry.get_saved_dataset(
                name=request.name,
                project=request.project,
                allow_cache=request.allow_cache,
            ),
            actions=[AuthzedAction.DESCRIBE],
        ).to_proto()

    def ListSavedDatasets(
        self, request: RegistryServer_pb2.ListSavedDatasetsRequest, context
    ):
        paginated_saved_datasets, pagination_metadata = apply_pagination_and_sorting(
            permitted_resources(
                resources=cast(
                    list[FeastObject],
                    self.proxied_registry.list_saved_datasets(
                        project=request.project,
                        allow_cache=request.allow_cache,
                        tags=dict(request.tags),
                    ),
                ),
                actions=AuthzedAction.DESCRIBE,
            ),
            pagination=request.pagination,
            sorting=request.sorting,
        )

        return RegistryServer_pb2.ListSavedDatasetsResponse(
            saved_datasets=[
                saved_dataset.to_proto() for saved_dataset in paginated_saved_datasets
            ],
            pagination=pagination_metadata,
        )

    def DeleteSavedDataset(
        self, request: RegistryServer_pb2.DeleteSavedDatasetRequest, context
    ):
        saved_dataset = self.proxied_registry.get_saved_dataset(
            name=request.name, project=request.project, allow_cache=False
        )
        assert_permissions(
            resource=saved_dataset,
            actions=[AuthzedAction.DELETE],
        )
        self.proxied_registry.delete_saved_dataset(
            name=request.name, project=request.project, commit=request.commit
        )

        return Empty()

    def ApplyValidationReference(
        self, request: RegistryServer_pb2.ApplyValidationReferenceRequest, context
    ):
        validation_reference = cast(
            ValidationReference,
            assert_permissions_to_update(
                resource=ValidationReference.from_proto(request.validation_reference),
                getter=self.proxied_registry.get_validation_reference,
                project=request.project,
            ),
        )
        self.proxied_registry.apply_validation_reference(
            validation_reference=validation_reference,
            project=request.project,
            commit=request.commit,
        )

        return Empty()

    def GetValidationReference(
        self, request: RegistryServer_pb2.GetValidationReferenceRequest, context
    ):
        return assert_permissions(
            self.proxied_registry.get_validation_reference(
                name=request.name,
                project=request.project,
                allow_cache=request.allow_cache,
            ),
            actions=[AuthzedAction.DESCRIBE],
        ).to_proto()

    def ListValidationReferences(
        self, request: RegistryServer_pb2.ListValidationReferencesRequest, context
    ):
        paginated_validation_references, pagination_metadata = (
            apply_pagination_and_sorting(
                permitted_resources(
                    resources=cast(
                        list[FeastObject],
                        self.proxied_registry.list_validation_references(
                            project=request.project,
                            allow_cache=request.allow_cache,
                            tags=dict(request.tags),
                        ),
                    ),
                    actions=AuthzedAction.DESCRIBE,
                ),
                pagination=request.pagination,
                sorting=request.sorting,
            )
        )

        return RegistryServer_pb2.ListValidationReferencesResponse(
            validation_references=[
                validation_reference.to_proto()
                for validation_reference in paginated_validation_references
            ],
            pagination=pagination_metadata,
        )

    def DeleteValidationReference(
        self, request: RegistryServer_pb2.DeleteValidationReferenceRequest, context
    ):
        validation_reference = self.proxied_registry.get_validation_reference(
            name=request.name, project=request.project, allow_cache=False
        )
        assert_permissions(
            resource=validation_reference,
            actions=[AuthzedAction.DELETE],
        )
        self.proxied_registry.delete_validation_reference(
            name=request.name, project=request.project, commit=request.commit
        )
        return Empty()

    def ListProjectMetadata(
        self, request: RegistryServer_pb2.ListProjectMetadataRequest, context
    ):
        return RegistryServer_pb2.ListProjectMetadataResponse(
            project_metadata=[
                project_metadata.to_proto()
                for project_metadata in self.proxied_registry.list_project_metadata(
                    project=request.project, allow_cache=request.allow_cache
                )
            ]
        )

    def ApplyMaterialization(
        self, request: RegistryServer_pb2.ApplyMaterializationRequest, context
    ):
        assert_permissions(
            resource=FeatureView.from_proto(request.feature_view),
            actions=[AuthzedAction.WRITE_ONLINE],
        )

        self.proxied_registry.apply_materialization(
            feature_view=FeatureView.from_proto(request.feature_view),
            project=request.project,
            start_date=datetime.fromtimestamp(
                request.start_date.seconds + request.start_date.nanos / 1e9,
                tz=timezone.utc,
            ),
            end_date=datetime.fromtimestamp(
                request.end_date.seconds + request.end_date.nanos / 1e9, tz=timezone.utc
            ),
            commit=request.commit,
        )
        return Empty()

    def UpdateInfra(self, request: RegistryServer_pb2.UpdateInfraRequest, context):
        self.proxied_registry.update_infra(
            infra=Infra.from_proto(request.infra),
            project=request.project,
            commit=request.commit,
        )
        return Empty()

    def GetInfra(self, request: RegistryServer_pb2.GetInfraRequest, context):
        return self.proxied_registry.get_infra(
            project=request.project, allow_cache=request.allow_cache
        ).to_proto()

    def ApplyPermission(
        self, request: RegistryServer_pb2.ApplyPermissionRequest, context
    ):
        permission = cast(
            Permission,
            assert_permissions_to_update(
                resource=Permission.from_proto(request.permission),
                getter=self.proxied_registry.get_permission,
                project=request.project,
            ),
        )
        self.proxied_registry.apply_permission(
            permission=permission,
            project=request.project,
            commit=request.commit,
        )
        return Empty()

    def GetPermission(self, request: RegistryServer_pb2.GetPermissionRequest, context):
        return assert_permissions(
            self.proxied_registry.get_permission(
                name=request.name,
                project=request.project,
                allow_cache=request.allow_cache,
            ),
            actions=[AuthzedAction.DESCRIBE],
        ).to_proto()

    def ListPermissions(
        self, request: RegistryServer_pb2.ListPermissionsRequest, context
    ):
        paginated_permissions, pagination_metadata = apply_pagination_and_sorting(
            permitted_resources(
                resources=cast(
                    list[FeastObject],
                    self.proxied_registry.list_permissions(
                        project=request.project,
                        allow_cache=request.allow_cache,
                        tags=dict(request.tags),
                    ),
                ),
                actions=AuthzedAction.DESCRIBE,
            ),
            pagination=request.pagination,
            sorting=request.sorting,
        )

        return RegistryServer_pb2.ListPermissionsResponse(
            permissions=[permission.to_proto() for permission in paginated_permissions],
            pagination=pagination_metadata,
        )

    def DeletePermission(
        self, request: RegistryServer_pb2.DeletePermissionRequest, context
    ):
        permission = self.proxied_registry.get_permission(
            name=request.name, project=request.project, allow_cache=False
        )
        assert_permissions(
            resource=permission,
            actions=[AuthzedAction.DELETE],
        )
        self.proxied_registry.delete_permission(
            name=request.name, project=request.project, commit=request.commit
        )

        return Empty()

    def ApplyProject(self, request: RegistryServer_pb2.ApplyProjectRequest, context):
        project = cast(
            Project,
            assert_permissions_to_update(
                resource=Project.from_proto(request.project),
                getter=self.proxied_registry.get_project,
                project=Project.from_proto(request.project).name,
            ),
        )
        self.proxied_registry.apply_project(
            project=project,
            commit=request.commit,
        )
        return Empty()

    def GetProject(self, request: RegistryServer_pb2.GetProjectRequest, context):
        return assert_permissions(
            self.proxied_registry.get_project(
                name=request.name,
                allow_cache=request.allow_cache,
            ),
            actions=[AuthzedAction.DESCRIBE],
        ).to_proto()

    def ListProjects(self, request: RegistryServer_pb2.ListProjectsRequest, context):
        paginated_projects, pagination_metadata = apply_pagination_and_sorting(
            permitted_resources(
                resources=cast(
                    list[FeastObject],
                    self.proxied_registry.list_projects(
                        allow_cache=request.allow_cache,
                        tags=dict(request.tags),
                    ),
                ),
                actions=AuthzedAction.DESCRIBE,
            ),
            pagination=request.pagination,
            sorting=request.sorting,
        )

        return RegistryServer_pb2.ListProjectsResponse(
            projects=[project.to_proto() for project in paginated_projects],
            pagination=pagination_metadata,
        )

    def DeleteProject(self, request: RegistryServer_pb2.DeleteProjectRequest, context):
        project = self.proxied_registry.get_project(
            name=request.name, allow_cache=False
        )
        assert_permissions(
            resource=project,
            actions=[AuthzedAction.DELETE],
        )
        self.proxied_registry.delete_project(name=request.name, commit=request.commit)

        return Empty()

    def GetRegistryLineage(
        self, request: RegistryServer_pb2.GetRegistryLineageRequest, context
    ):
        direct_relationships, indirect_relationships = (
            self.proxied_registry.get_registry_lineage(
                project=request.project,
                allow_cache=request.allow_cache,
                filter_object_type=request.filter_object_type,
                filter_object_name=request.filter_object_name,
            )
        )

        paginated_relationships, relationships_pagination = (
            apply_pagination_and_sorting(
                direct_relationships,
                pagination=request.pagination,
                sorting=request.sorting,
            )
        )

        paginated_indirect_relationships, indirect_relationships_pagination = (
            apply_pagination_and_sorting(
                indirect_relationships,
                pagination=request.pagination,
                sorting=request.sorting,
            )
        )

        return RegistryServer_pb2.GetRegistryLineageResponse(
            relationships=[rel.to_proto() for rel in paginated_relationships],
            indirect_relationships=[
                rel.to_proto() for rel in paginated_indirect_relationships
            ],
            relationships_pagination=relationships_pagination,
            indirect_relationships_pagination=indirect_relationships_pagination,
        )

    def GetObjectRelationships(
        self, request: RegistryServer_pb2.GetObjectRelationshipsRequest, context
    ):
        """Get relationships for a specific object."""
        relationships = self.proxied_registry.get_object_relationships(
            project=request.project,
            object_type=request.object_type,
            object_name=request.object_name,
            include_indirect=request.include_indirect,
            allow_cache=request.allow_cache,
        )

        paginated_relationships, pagination_metadata = apply_pagination_and_sorting(
            relationships,
            pagination=request.pagination,
            sorting=request.sorting,
        )

        return RegistryServer_pb2.GetObjectRelationshipsResponse(
            relationships=[rel.to_proto() for rel in paginated_relationships],
            pagination=pagination_metadata,
        )

    def Commit(self, request, context):
        self.proxied_registry.commit()
        return Empty()

    def Refresh(self, request, context):
        self.proxied_registry.refresh(request.project)
        return Empty()

    def Proto(self, request, context):
        return self.proxied_registry.proto()

    def ListFeatures(self, request: RegistryServer_pb2.ListFeaturesRequest, context):
        """
        List all features in the registry, optionally filtered by project, feature_view, or name.
        """
        allow_cache = request.allow_cache if hasattr(request, "allow_cache") else True
        feature_views = self.proxied_registry.list_all_feature_views(
            project=request.project,
            allow_cache=allow_cache,
        )
        permitted_fvs = permitted_resources(
            resources=cast(list[FeastObject], feature_views),
            actions=AuthzedAction.DESCRIBE,
        )
        features = []
        for fv in permitted_fvs:
            fv_name = getattr(fv, "name", None)
            for feature in getattr(fv, "features", []):
                if request.feature_view and fv_name != request.feature_view:
                    continue
                if request.name and feature.name != request.name:
                    continue
                # Get owner and timestamps from feature view
                owner = ""
                created_timestamp = None
                last_updated_timestamp = None

                if hasattr(fv, "spec") and hasattr(fv.spec, "owner"):
                    owner = getattr(fv.spec, "owner", "")

                if hasattr(fv, "meta"):
                    if hasattr(fv.meta, "created_timestamp"):
                        created_timestamp = fv.meta.created_timestamp
                    if hasattr(fv.meta, "last_updated_timestamp"):
                        last_updated_timestamp = fv.meta.last_updated_timestamp

                features.append(
                    Feature(
                        name=feature.name,
                        feature_view=fv_name if fv_name is not None else "",
                        type=str(feature.dtype)
                        if hasattr(feature, "dtype")
                        else str(feature.valueType),
                        description=getattr(feature, "description", ""),
                        owner=owner,
                        created_timestamp=created_timestamp,
                        last_updated_timestamp=last_updated_timestamp,
                        tags=getattr(feature, "tags", {}),
                    )
                )
        paginated_features, pagination_metadata = apply_pagination_and_sorting(
            features,
            pagination=request.pagination if request.HasField("pagination") else None,
            sorting=request.sorting if request.HasField("sorting") else None,
        )
        return ListFeaturesResponse(
            features=paginated_features, pagination=pagination_metadata
        )

    def GetFeature(self, request: RegistryServer_pb2.GetFeatureRequest, context):
        """
        Get a single feature by project, feature_view, and name.
        """
        allow_cache = getattr(request, "allow_cache", True)
        feature_views = self.proxied_registry.list_all_feature_views(
            project=request.project,
            allow_cache=allow_cache,
        )
        permitted_fvs = permitted_resources(
            resources=cast(list[FeastObject], feature_views),
            actions=AuthzedAction.DESCRIBE,
        )
        for fv in permitted_fvs:
            fv_name = getattr(fv, "name", None)
            for feature in getattr(fv, "features", []):
                if fv_name == request.feature_view and feature.name == request.name:
                    # Get owner and timestamps from feature view
                    owner = ""
                    created_timestamp = None
                    last_updated_timestamp = None

                    if hasattr(fv, "spec") and hasattr(fv.spec, "owner"):
                        owner = getattr(fv.spec, "owner", "")

                    if hasattr(fv, "meta"):
                        if hasattr(fv.meta, "created_timestamp"):
                            created_timestamp = fv.meta.created_timestamp
                        if hasattr(fv.meta, "last_updated_timestamp"):
                            last_updated_timestamp = fv.meta.last_updated_timestamp

                    return Feature(
                        name=feature.name,
                        feature_view=fv_name if fv_name is not None else "",
                        type=str(feature.dtype)
                        if hasattr(feature, "dtype")
                        else str(feature.valueType),
                        description=getattr(feature, "description", ""),
                        owner=owner,
                        created_timestamp=created_timestamp,
                        last_updated_timestamp=last_updated_timestamp,
                        tags=getattr(feature, "tags", {}),
                    )
        raise FeastObjectNotFoundException(
            f"Feature {request.name} not found in feature view {request.feature_view} in project {request.project}"
        )


def start_server(
    store: FeatureStore,
    port: int,
    wait_for_termination: bool = True,
    tls_key_path: str = "",
    tls_cert_path: str = "",
):
    auth_manager_type = str_to_auth_manager_type(store.config.auth_config.type)
    init_security_manager(auth_type=auth_manager_type, fs=store)
    init_auth_manager(
        auth_type=auth_manager_type,
        server_type=ServerType.GRPC,
        auth_config=store.config.auth_config,
    )

    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=10),
        interceptors=_grpc_interceptors(auth_manager_type),
    )
    RegistryServer_pb2_grpc.add_RegistryServerServicer_to_server(
        RegistryServer(store.registry), server
    )
    # Add health check service to server
    health_servicer = health.HealthServicer()
    health_pb2_grpc.add_HealthServicer_to_server(health_servicer, server)
    health_servicer.set("", health_pb2.HealthCheckResponse.SERVING)

    service_names_available_for_reflection = (
        RegistryServer_pb2.DESCRIPTOR.services_by_name["RegistryServer"].full_name,
        health_pb2.DESCRIPTOR.services_by_name["Health"].full_name,
        reflection.SERVICE_NAME,
    )
    reflection.enable_server_reflection(service_names_available_for_reflection, server)

    if tls_cert_path and tls_key_path:
        with (
            open(tls_cert_path, "rb") as cert_file,
            open(tls_key_path, "rb") as key_file,
        ):
            certificate_chain = cert_file.read()
            private_key = key_file.read()
            server_credentials = grpc.ssl_server_credentials(
                ((private_key, certificate_chain),)
            )
        logger.info("Starting grpc registry server in TLS(SSL) mode")
        server.add_secure_port(f"[::]:{port}", server_credentials)
    else:
        logger.info("Starting grpc registry server in non-TLS(SSL) mode")
        server.add_insecure_port(f"[::]:{port}")
    server.start()
    if wait_for_termination:
        logger.info(f"Grpc server started at localhost:{port}")
        server.wait_for_termination()
    else:
        return server


def _grpc_interceptors(
    auth_type: AuthManagerType,
) -> Optional[list[grpc.ServerInterceptor]]:
    """
    A list of the interceptors for the registry server.

    Args:
        auth_type: The type of authorization manager, from the feature store configuration.

    Returns:
        list[grpc.ServerInterceptor]: Optional list of interceptors. If the authorization type is set to `NONE`, it returns `None`.
    """
    if auth_type == AuthManagerType.NONE:
        return [ErrorInterceptor()]

    return [AuthInterceptor(), ErrorInterceptor()]



"""
Registry lineage generation for Feast objects.

This module provides functionality to generate relationship graphs between
Feast objects (entities, feature views, data sources, feature services)
for lineage visualization.
"""

from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Tuple

from feast.protos.feast.core.Registry_pb2 import Registry


class FeastObjectType(Enum):
    DATA_SOURCE = "dataSource"
    ENTITY = "entity"
    FEATURE_VIEW = "featureView"
    FEATURE_SERVICE = "featureService"
    FEATURE = "feature"


@dataclass
class EntityReference:
    type: FeastObjectType
    name: str

    def to_proto(self):
        try:
            from feast.protos.feast.registry.RegistryServer_pb2 import (
                EntityReference as EntityReferenceProto,
            )

            return EntityReferenceProto(type=self.type.value, name=self.name)
        except ImportError:
            return {"type": self.type.value, "name": self.name}


@dataclass
class EntityRelation:
    source: EntityReference
    target: EntityReference

    def to_proto(self):
        try:
            from feast.protos.feast.registry.RegistryServer_pb2 import (
                EntityRelation as EntityRelationProto,
            )

            return EntityRelationProto(
                source=self.source.to_proto(), target=self.target.to_proto()
            )
        except ImportError:
            # Fallback to dict if protobuf not generated yet
            return {"source": self.source.to_proto(), "target": self.target.to_proto()}


class RegistryLineageGenerator:
    """
    Generates lineage relationships between Feast objects.
    """

    def generate_lineage(
        self, registry: Registry
    ) -> Tuple[List[EntityRelation], List[EntityRelation]]:
        """
        Generate both direct and indirect relationships from registry objects.
        Args:
            registry: The registry protobuf containing all objects
        Returns:
            Tuple of (direct_relationships, indirect_relationships)
        """
        direct_relationships = self._parse_direct_relationships(registry)
        indirect_relationships = self._parse_indirect_relationships(
            direct_relationships, registry
        )

        return direct_relationships, indirect_relationships

    def _parse_direct_relationships(self, registry: Registry) -> List[EntityRelation]:
        """Parse direct relationships between objects."""
        relationships = []

        # FeatureService -> FeatureView relationships
        for feature_service in registry.feature_services:
            if (
                hasattr(feature_service, "spec")
                and feature_service.spec
                and feature_service.spec.features
            ):
                for feature in feature_service.spec.features:
                    rel = EntityRelation(
                        source=EntityReference(
                            FeastObjectType.FEATURE_VIEW, feature.feature_view_name
                        ),
                        target=EntityReference(
                            FeastObjectType.FEATURE_SERVICE,
                            feature_service.spec.name,
                        ),
                    )
                    relationships.append(rel)

        # Entity -> FeatureView and DataSource -> FeatureView relationships
        for feature_view in registry.feature_views:
            if hasattr(feature_view, "spec") and feature_view.spec:
                # Entity relationships
                if hasattr(feature_view.spec, "entities"):
                    for entity_name in feature_view.spec.entities:
                        rel = EntityRelation(
                            source=EntityReference(FeastObjectType.ENTITY, entity_name),
                            target=EntityReference(
                                FeastObjectType.FEATURE_VIEW, feature_view.spec.name
                            ),
                        )
                        relationships.append(rel)

                # Feature -> FeatureView relationships
                if hasattr(feature_view.spec, "features"):
                    for feature in feature_view.spec.features:
                        rel = EntityRelation(
                            source=EntityReference(
                                FeastObjectType.FEATURE, feature.name
                            ),
                            target=EntityReference(
                                FeastObjectType.FEATURE_VIEW, feature_view.spec.name
                            ),
                        )
                        relationships.append(rel)

                # Batch source relationship
                if (
                    hasattr(feature_view.spec, "batch_source")
                    and feature_view.spec.batch_source
                ):
                    # Try to get the data source name
                    data_source_name = None
                    if (
                        hasattr(feature_view.spec.batch_source, "name")
                        and feature_view.spec.batch_source.name
                    ):
                        data_source_name = feature_view.spec.batch_source.name
                    elif (
                        hasattr(feature_view.spec.batch_source, "table")
                        and feature_view.spec.batch_source.table
                    ):
                        # Fallback to table name for unnamed data sources
                        data_source_name = (
                            f"table:{feature_view.spec.batch_source.table}"
                        )
                    elif (
                        hasattr(feature_view.spec.batch_source, "path")
                        and feature_view.spec.batch_source.path
                    ):
                        # Fallback to path for file-based sources
                        data_source_name = f"path:{feature_view.spec.batch_source.path}"
                    else:
                        # Use a generic identifier
                        data_source_name = f"unnamed_source_{hash(str(feature_view.spec.batch_source))}"

                    if data_source_name:
                        relationships.append(
                            EntityRelation(
                                source=EntityReference(
                                    FeastObjectType.DATA_SOURCE, data_source_name
                                ),
                                target=EntityReference(
                                    FeastObjectType.FEATURE_VIEW, feature_view.spec.name
                                ),
                            )
                        )

        # OnDemand FeatureView: Feature -> OnDemandFeatureView relationships
        for odfv in registry.on_demand_feature_views:
            if hasattr(odfv, "spec") and odfv.spec:
                # Entity relationships
                if hasattr(odfv.spec, "entities"):
                    for entity_name in odfv.spec.entities:
                        rel = EntityRelation(
                            source=EntityReference(FeastObjectType.ENTITY, entity_name),
                            target=EntityReference(
                                FeastObjectType.FEATURE_VIEW, odfv.spec.name
                            ),
                        )
                        relationships.append(rel)

                # Feature -> OnDemandFeatureView relationships
                if hasattr(odfv.spec, "features"):
                    for feature in odfv.spec.features:
                        relationships.append(
                            EntityRelation(
                                source=EntityReference(
                                    FeastObjectType.FEATURE, feature.name
                                ),
                                target=EntityReference(
                                    FeastObjectType.FEATURE_VIEW, odfv.spec.name
                                ),
                            )
                        )

        # OnDemand FeatureView relationships
        for odfv in registry.on_demand_feature_views:
            if (
                hasattr(odfv, "spec")
                and odfv.spec
                and hasattr(odfv.spec, "sources")
                and odfv.spec.sources
            ):
                # Handle protobuf map structure
                if hasattr(odfv.spec.sources, "items"):
                    source_items = odfv.spec.sources.items()
                else:
                    # Fallback for different protobuf representations
                    source_items = [(k, v) for k, v in enumerate(odfv.spec.sources)]

                for source_name, source in source_items:
                    if (
                        hasattr(source, "request_data_source")
                        and source.request_data_source
                    ):
                        if hasattr(source.request_data_source, "name"):
                            relationships.append(
                                EntityRelation(
                                    source=EntityReference(
                                        FeastObjectType.DATA_SOURCE,
                                        source.request_data_source.name,
                                    ),
                                    target=EntityReference(
                                        FeastObjectType.FEATURE_VIEW, odfv.spec.name
                                    ),
                                )
                            )
                    elif (
                        hasattr(source, "feature_view_projection")
                        and source.feature_view_projection
                    ):
                        # Find the source feature view's batch source
                        if hasattr(source.feature_view_projection, "feature_view_name"):
                            source_fv = next(
                                (
                                    fv
                                    for fv in registry.feature_views
                                    if hasattr(fv, "spec")
                                    and fv.spec
                                    and hasattr(fv.spec, "name")
                                    and fv.spec.name
                                    == source.feature_view_projection.feature_view_name
                                ),
                                None,
                            )
                            if (
                                source_fv
                                and hasattr(source_fv, "spec")
                                and source_fv.spec
                                and hasattr(source_fv.spec, "batch_source")
                                and source_fv.spec.batch_source
                                and hasattr(source_fv.spec.batch_source, "name")
                            ):
                                relationships.append(
                                    EntityRelation(
                                        source=EntityReference(
                                            FeastObjectType.DATA_SOURCE,
                                            source_fv.spec.batch_source.name,
                                        ),
                                        target=EntityReference(
                                            FeastObjectType.FEATURE_VIEW, odfv.spec.name
                                        ),
                                    )
                                )

        # Stream FeatureView relationships
        for sfv in registry.stream_feature_views:
            if hasattr(sfv, "spec") and sfv.spec:
                # Stream source
                if (
                    hasattr(sfv.spec, "stream_source")
                    and sfv.spec.stream_source
                    and hasattr(sfv.spec.stream_source, "name")
                    and sfv.spec.stream_source.name
                ):
                    relationships.append(
                        EntityRelation(
                            source=EntityReference(
                                FeastObjectType.DATA_SOURCE, sfv.spec.stream_source.name
                            ),
                            target=EntityReference(
                                FeastObjectType.FEATURE_VIEW, sfv.spec.name
                            ),
                        )
                    )

                # Batch source
                if (
                    hasattr(sfv.spec, "batch_source")
                    and sfv.spec.batch_source
                    and hasattr(sfv.spec.batch_source, "name")
                    and sfv.spec.batch_source.name
                ):
                    relationships.append(
                        EntityRelation(
                            source=EntityReference(
                                FeastObjectType.DATA_SOURCE, sfv.spec.batch_source.name
                            ),
                            target=EntityReference(
                                FeastObjectType.FEATURE_VIEW, sfv.spec.name
                            ),
                        )
                    )

        return relationships

    def _parse_indirect_relationships(
        self, direct_relationships: List[EntityRelation], registry: Registry
    ) -> List[EntityRelation]:
        """Parse indirect relationships (transitive relationships through feature views)."""
        indirect_relationships = []

        # Create Entity -> FeatureService and DataSource -> FeatureService relationships
        for feature_service in registry.feature_services:
            if (
                hasattr(feature_service, "spec")
                and feature_service.spec
                and hasattr(feature_service.spec, "features")
                and feature_service.spec.features
            ):
                for feature in feature_service.spec.features:
                    if hasattr(feature, "feature_view_name"):
                        # Find all relationships that target this feature view
                        related_sources = [
                            rel.source
                            for rel in direct_relationships
                            if rel.target.name == feature.feature_view_name
                            and rel.target.type == FeastObjectType.FEATURE_VIEW
                        ]

                        # Create indirect relationships to the feature service
                        for source in related_sources:
                            indirect_relationships.append(
                                EntityRelation(
                                    source=source,
                                    target=EntityReference(
                                        FeastObjectType.FEATURE_SERVICE,
                                        feature_service.spec.name,
                                    ),
                                )
                            )

        # Create Entity -> DataSource relationships (through feature views)
        # Build a map of feature view -> data sources
        feature_view_to_data_sources: Dict[str, List[str]] = {}
        for rel in direct_relationships:
            if (
                rel.source.type == FeastObjectType.DATA_SOURCE
                and rel.target.type == FeastObjectType.FEATURE_VIEW
            ):
                if rel.target.name not in feature_view_to_data_sources:
                    feature_view_to_data_sources[rel.target.name] = []
                feature_view_to_data_sources[rel.target.name].append(rel.source.name)

        # For each Entity -> FeatureView relationship, create Entity -> DataSource relationships
        for rel in direct_relationships:
            if (
                rel.source.type == FeastObjectType.ENTITY
                and rel.target.type == FeastObjectType.FEATURE_VIEW
            ):
                # Find data sources that this feature view uses
                if rel.target.name in feature_view_to_data_sources:
                    for data_source_name in feature_view_to_data_sources[
                        rel.target.name
                    ]:
                        indirect_relationships.append(
                            EntityRelation(
                                source=rel.source,  # The entity
                                target=EntityReference(
                                    FeastObjectType.DATA_SOURCE,
                                    data_source_name,
                                ),
                            )
                        )

        return indirect_relationships

    def get_object_relationships(
        self,
        registry: Registry,
        object_type: str,
        object_name: str,
        include_indirect: bool = False,
    ) -> List[EntityRelation]:
        """
        Get all relationships for a specific object.
        Args:
            registry: The registry protobuf
            object_type: Type of the object (dataSource, entity, featureView, featureService)
            object_name: Name of the object
            include_indirect: Whether to include indirect relationships
        Returns:
            List of relationships involving the specified object
        """
        direct_relationships, indirect_relationships = self.generate_lineage(registry)

        all_relationships = direct_relationships[:]
        if include_indirect:
            all_relationships.extend(indirect_relationships)

        # Filter relationships involving the specified object
        filtered_relationships = []
        target_type = FeastObjectType(object_type)

        for rel in all_relationships:
            if (rel.source.type == target_type and rel.source.name == object_name) or (
                rel.target.type == target_type and rel.target.name == object_name
            ):
                filtered_relationships.append(rel)

        return filtered_relationships

    def get_object_lineage_graph(
        self, registry: Registry, object_type: str, object_name: str, depth: int = 2
    ) -> Dict:
        """
        Get a complete lineage graph for an object up to specified depth.
        This can be used for more complex lineage queries and visualization.
        """
        direct_relationships, indirect_relationships = self.generate_lineage(registry)
        all_relationships = direct_relationships + indirect_relationships

        # Build adjacency graph
        graph: Dict[str, List[str]] = {}
        for rel in all_relationships:
            source_key = f"{rel.source.type.value}:{rel.source.name}"
            target_key = f"{rel.target.type.value}:{rel.target.name}"

            if source_key not in graph:
                graph[source_key] = []
            graph[source_key].append(target_key)

        # Perform BFS to get subgraph up to specified depth
        start_key = f"{object_type}:{object_name}"
        visited = set()
        result_nodes = set()
        result_edges = []

        def bfs(current_key, current_depth):
            if current_depth > depth or current_key in visited:
                return

            visited.add(current_key)
            result_nodes.add(current_key)

            if current_key in graph:
                for neighbor in graph[current_key]:
                    result_edges.append((current_key, neighbor))
                    result_nodes.add(neighbor)
                    bfs(neighbor, current_depth + 1)

        bfs(start_key, 0)

        return {"nodes": list(result_nodes), "edges": result_edges}



import pyarrow.flight as fl

from feast.permissions.auth_model import AuthConfig
from feast.permissions.client.client_auth_token import get_auth_token


class FlightBearerTokenInterceptor(fl.ClientMiddleware):
    def __init__(self, auth_config: AuthConfig):
        super().__init__()
        self.auth_config = auth_config

    def call_completed(self, exception):
        pass

    def received_headers(self, headers):
        pass

    def sending_headers(self):
        access_token = get_auth_token(self.auth_config)
        return {b"authorization": b"Bearer " + access_token.encode("utf-8")}


class FlightAuthInterceptorFactory(fl.ClientMiddlewareFactory):
    def __init__(self, auth_config: AuthConfig):
        super().__init__()
        self.auth_config = auth_config

    def start_call(self, info):
        return FlightBearerTokenInterceptor(self.auth_config)



# MCP (Model Context Protocol) server implementations for Feast

from .mcp_config import McpFeatureServerConfig
from .mcp_server import add_mcp_support_to_app

__all__ = [
    "McpFeatureServerConfig",
    "add_mcp_support_to_app",
]



import click

from feast.repo_operations import create_feature_store, registry_dump


@click.command()
@click.option(
    "--host",
    "-h",
    type=click.STRING,
    default="0.0.0.0",
    show_default=True,
    help="Specify a host for the server",
)
@click.option(
    "--port",
    "-p",
    type=click.INT,
    default=8888,
    show_default=True,
    help="Specify a port for the server",
)
@click.option(
    "--registry_ttl_sec",
    "-r",
    help="Number of seconds after which the registry is refreshed",
    type=click.INT,
    default=5,
    show_default=True,
)
@click.option(
    "--root_path",
    help="Provide root path to make the UI working behind proxy",
    type=click.STRING,
    default="",
)
@click.option(
    "--key",
    "-k",
    "tls_key_path",
    type=click.STRING,
    default="",
    show_default=False,
    help="path to TLS(SSL) certificate private key. You need to pass --cert arg as well to start server in TLS mode",
)
@click.option(
    "--cert",
    "-c",
    "tls_cert_path",
    type=click.STRING,
    default="",
    show_default=False,
    help="path to TLS(SSL) certificate public key. You need to pass --key arg as well to start server in TLS mode",
)
@click.pass_context
def ui(
    ctx: click.Context,
    host: str,
    port: int,
    registry_ttl_sec: int,
    root_path: str = "",
    tls_key_path: str = "",
    tls_cert_path: str = "",
):
    """
    Shows the Feast UI over the current directory
    """
    if (tls_key_path and not tls_cert_path) or (not tls_key_path and tls_cert_path):
        raise click.BadParameter(
            "Please configure --key and --cert args to start the feature server in SSL mode."
        )
    store = create_feature_store(ctx)
    # Pass in the registry_dump method to get around a circular dependency
    store.serve_ui(
        host=host,
        port=port,
        get_registry_dump=registry_dump,
        registry_ttl_sec=registry_ttl_sec,
        root_path=root_path,
        tls_key_path=tls_key_path,
        tls_cert_path=tls_cert_path,
    )



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Copyright 2018 The Feast Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import feast.types.Value_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _FieldStatus:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _FieldStatusEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_FieldStatus.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    INVALID: _FieldStatus.ValueType  # 0
    """Status is unset for this field."""
    PRESENT: _FieldStatus.ValueType  # 1
    """Field value is present for this field and age is within max age."""
    NULL_VALUE: _FieldStatus.ValueType  # 2
    """Values could be found for entity key and age is within max age, but
    this field value is not assigned a value on ingestion into feast.
    """
    NOT_FOUND: _FieldStatus.ValueType  # 3
    """Entity key did not return any values as they do not exist in Feast.
    This could suggest that the feature values have not yet been ingested
    into feast or the ingestion failed.
    """
    OUTSIDE_MAX_AGE: _FieldStatus.ValueType  # 4
    """Values could be found for entity key, but field values are outside the maximum
    allowable range.
    """

class FieldStatus(_FieldStatus, metaclass=_FieldStatusEnumTypeWrapper): ...

INVALID: FieldStatus.ValueType  # 0
"""Status is unset for this field."""
PRESENT: FieldStatus.ValueType  # 1
"""Field value is present for this field and age is within max age."""
NULL_VALUE: FieldStatus.ValueType  # 2
"""Values could be found for entity key and age is within max age, but
this field value is not assigned a value on ingestion into feast.
"""
NOT_FOUND: FieldStatus.ValueType  # 3
"""Entity key did not return any values as they do not exist in Feast.
This could suggest that the feature values have not yet been ingested
into feast or the ingestion failed.
"""
OUTSIDE_MAX_AGE: FieldStatus.ValueType  # 4
"""Values could be found for entity key, but field values are outside the maximum
allowable range.
"""
global___FieldStatus = FieldStatus

class GetFeastServingInfoRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___GetFeastServingInfoRequest = GetFeastServingInfoRequest

class GetFeastServingInfoResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VERSION_FIELD_NUMBER: builtins.int
    version: builtins.str
    """Feast version of this serving deployment."""
    def __init__(
        self,
        *,
        version: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["version", b"version"]) -> None: ...

global___GetFeastServingInfoResponse = GetFeastServingInfoResponse

class FeatureReferenceV2(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FEATURE_VIEW_NAME_FIELD_NUMBER: builtins.int
    FEATURE_NAME_FIELD_NUMBER: builtins.int
    feature_view_name: builtins.str
    """Name of the Feature View to retrieve the feature from."""
    feature_name: builtins.str
    """Name of the Feature to retrieve the feature from."""
    def __init__(
        self,
        *,
        feature_view_name: builtins.str = ...,
        feature_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["feature_name", b"feature_name", "feature_view_name", b"feature_view_name"]) -> None: ...

global___FeatureReferenceV2 = FeatureReferenceV2

class GetOnlineFeaturesRequestV2(google.protobuf.message.Message):
    """ToDo (oleksii): remove this message (since it's not used) and move EntityRow on package level"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class EntityRow(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class FieldsEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            @property
            def value(self) -> feast.types.Value_pb2.Value: ...
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: feast.types.Value_pb2.Value | None = ...,
            ) -> None: ...
            def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

        TIMESTAMP_FIELD_NUMBER: builtins.int
        FIELDS_FIELD_NUMBER: builtins.int
        @property
        def timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
            """Request timestamp of this row. This value will be used,
            together with maxAge, to determine feature staleness.
            """
        @property
        def fields(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, feast.types.Value_pb2.Value]:
            """Map containing mapping of entity name to entity value."""
        def __init__(
            self,
            *,
            timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
            fields: collections.abc.Mapping[builtins.str, feast.types.Value_pb2.Value] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["timestamp", b"timestamp"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["fields", b"fields", "timestamp", b"timestamp"]) -> None: ...

    FEATURES_FIELD_NUMBER: builtins.int
    ENTITY_ROWS_FIELD_NUMBER: builtins.int
    PROJECT_FIELD_NUMBER: builtins.int
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FeatureReferenceV2]:
        """List of features that are being retrieved"""
    @property
    def entity_rows(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___GetOnlineFeaturesRequestV2.EntityRow]:
        """List of entity rows, containing entity id and timestamp data.
        Used during retrieval of feature rows and for joining feature
        rows into a final dataset
        """
    project: builtins.str
    """Optional field to specify project name override. If specified, uses the
    given project for retrieval. Overrides the projects specified in
    Feature References if both are specified.
    """
    def __init__(
        self,
        *,
        features: collections.abc.Iterable[global___FeatureReferenceV2] | None = ...,
        entity_rows: collections.abc.Iterable[global___GetOnlineFeaturesRequestV2.EntityRow] | None = ...,
        project: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["entity_rows", b"entity_rows", "features", b"features", "project", b"project"]) -> None: ...

global___GetOnlineFeaturesRequestV2 = GetOnlineFeaturesRequestV2

class FeatureList(google.protobuf.message.Message):
    """In JSON "val" field can be omitted"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VAL_FIELD_NUMBER: builtins.int
    @property
    def val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        val: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["val", b"val"]) -> None: ...

global___FeatureList = FeatureList

class GetOnlineFeaturesRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class EntitiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> feast.types.Value_pb2.RepeatedValue: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: feast.types.Value_pb2.RepeatedValue | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    class RequestContextEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> feast.types.Value_pb2.RepeatedValue: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: feast.types.Value_pb2.RepeatedValue | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    FEATURE_SERVICE_FIELD_NUMBER: builtins.int
    FEATURES_FIELD_NUMBER: builtins.int
    ENTITIES_FIELD_NUMBER: builtins.int
    FULL_FEATURE_NAMES_FIELD_NUMBER: builtins.int
    REQUEST_CONTEXT_FIELD_NUMBER: builtins.int
    feature_service: builtins.str
    @property
    def features(self) -> global___FeatureList: ...
    @property
    def entities(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, feast.types.Value_pb2.RepeatedValue]:
        """The entity data is specified in a columnar format
        A map of entity name -> list of values
        """
    full_feature_names: builtins.bool
    @property
    def request_context(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, feast.types.Value_pb2.RepeatedValue]:
        """Context for OnDemand Feature Transformation
        (was moved to dedicated parameter to avoid unnecessary separation logic on serving side)
        A map of variable name -> list of values
        """
    def __init__(
        self,
        *,
        feature_service: builtins.str = ...,
        features: global___FeatureList | None = ...,
        entities: collections.abc.Mapping[builtins.str, feast.types.Value_pb2.RepeatedValue] | None = ...,
        full_feature_names: builtins.bool = ...,
        request_context: collections.abc.Mapping[builtins.str, feast.types.Value_pb2.RepeatedValue] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["feature_service", b"feature_service", "features", b"features", "kind", b"kind"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["entities", b"entities", "feature_service", b"feature_service", "features", b"features", "full_feature_names", b"full_feature_names", "kind", b"kind", "request_context", b"request_context"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["kind", b"kind"]) -> typing_extensions.Literal["feature_service", "features"] | None: ...

global___GetOnlineFeaturesRequest = GetOnlineFeaturesRequest

class GetOnlineFeaturesResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class FeatureVector(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        VALUES_FIELD_NUMBER: builtins.int
        STATUSES_FIELD_NUMBER: builtins.int
        EVENT_TIMESTAMPS_FIELD_NUMBER: builtins.int
        @property
        def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.types.Value_pb2.Value]: ...
        @property
        def statuses(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[global___FieldStatus.ValueType]: ...
        @property
        def event_timestamps(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[google.protobuf.timestamp_pb2.Timestamp]: ...
        def __init__(
            self,
            *,
            values: collections.abc.Iterable[feast.types.Value_pb2.Value] | None = ...,
            statuses: collections.abc.Iterable[global___FieldStatus.ValueType] | None = ...,
            event_timestamps: collections.abc.Iterable[google.protobuf.timestamp_pb2.Timestamp] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["event_timestamps", b"event_timestamps", "statuses", b"statuses", "values", b"values"]) -> None: ...

    METADATA_FIELD_NUMBER: builtins.int
    RESULTS_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    @property
    def metadata(self) -> global___GetOnlineFeaturesResponseMetadata: ...
    @property
    def results(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___GetOnlineFeaturesResponse.FeatureVector]:
        """Length of "results" array should match length of requested features.
        We also preserve the same order of features here as in metadata.feature_names
        """
    status: builtins.bool
    def __init__(
        self,
        *,
        metadata: global___GetOnlineFeaturesResponseMetadata | None = ...,
        results: collections.abc.Iterable[global___GetOnlineFeaturesResponse.FeatureVector] | None = ...,
        status: builtins.bool = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["metadata", b"metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["metadata", b"metadata", "results", b"results", "status", b"status"]) -> None: ...

global___GetOnlineFeaturesResponse = GetOnlineFeaturesResponse

class GetOnlineFeaturesResponseMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FEATURE_NAMES_FIELD_NUMBER: builtins.int
    @property
    def feature_names(self) -> global___FeatureList: ...
    def __init__(
        self,
        *,
        feature_names: global___FeatureList | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["feature_names", b"feature_names"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["feature_names", b"feature_names"]) -> None: ...

global___GetOnlineFeaturesResponseMetadata = GetOnlineFeaturesResponseMetadata



from typing import Any

import click
import yaml
from bigtree import Node
from colorama import Fore, Style

from feast import (
    BatchFeatureView,
    Entity,
    FeatureService,
    FeatureView,
    OnDemandFeatureView,
    StreamFeatureView,
    utils,
)
from feast.cli import cli_utils
from feast.cli.cli_options import tagsOption
from feast.data_source import DataSource
from feast.errors import FeastObjectNotFoundException
from feast.permissions.policy import RoleBasedPolicy
from feast.repo_operations import create_feature_store
from feast.saved_dataset import SavedDataset, ValidationReference


@click.group(name="permissions")
def feast_permissions_cmd():
    """
    Access permissions
    """
    pass


@feast_permissions_cmd.command(name="list")
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Print the resources matching each configured permission",
)
@tagsOption
@click.pass_context
def feast_permissions_list_command(ctx: click.Context, verbose: bool, tags: list[str]):
    from tabulate import tabulate

    table: list[Any] = []
    tags_filter = utils.tags_list_to_dict(tags)

    store = create_feature_store(ctx)

    permissions = store.list_permissions(tags=tags_filter)

    root_node = Node("permissions")
    roles: set[str] = set()

    for p in permissions:
        policy = p.policy
        if not verbose:
            cli_utils.handle_not_verbose_permissions_command(p, policy, table)
        else:
            if isinstance(policy, RoleBasedPolicy) and len(policy.get_roles()) > 0:
                roles = set(policy.get_roles())
                permission_node = Node(
                    p.name + " " + str(list(roles)), parent=root_node
                )
            else:
                permission_node = Node(p.name, parent=root_node)

            for feast_type in p.types:
                if feast_type in [
                    FeatureView,
                    OnDemandFeatureView,
                    BatchFeatureView,
                    StreamFeatureView,
                ]:
                    cli_utils.handle_fv_verbose_permissions_command(
                        feast_type,  # type: ignore[arg-type]
                        p,
                        permission_node,
                        store,
                        tags_filter,
                    )
                elif feast_type == Entity:
                    cli_utils.handle_entity_verbose_permissions_command(
                        feast_type,  # type: ignore[arg-type]
                        p,
                        permission_node,
                        store,
                        tags_filter,
                    )
                elif feast_type == FeatureService:
                    cli_utils.handle_fs_verbose_permissions_command(
                        feast_type,  # type: ignore[arg-type]
                        p,
                        permission_node,
                        store,
                        tags_filter,
                    )
                elif feast_type == DataSource:
                    cli_utils.handle_ds_verbose_permissions_command(
                        feast_type,  # type: ignore[arg-type]
                        p,
                        permission_node,
                        store,
                        tags_filter,
                    )
                elif feast_type == ValidationReference:
                    cli_utils.handle_vr_verbose_permissions_command(
                        feast_type,  # type: ignore[arg-type]
                        p,
                        permission_node,
                        store,
                        tags_filter,
                    )
                elif feast_type == SavedDataset:
                    cli_utils.handle_sd_verbose_permissions_command(
                        feast_type,  # type: ignore[arg-type]
                        p,
                        permission_node,
                        store,
                        tags_filter,
                    )

    if not verbose:
        print(
            tabulate(
                table,
                headers=[
                    "NAME",
                    "TYPES",
                    "NAME_PATTERNS",
                    "ACTIONS",
                    "ROLES",
                    "REQUIRED_TAGS",
                ],
                tablefmt="plain",
            )
        )
    else:
        cli_utils.print_permission_verbose_example()

        print("Permissions:")
        print("")
        root_node.show()


@feast_permissions_cmd.command("describe")
@click.argument("name", type=click.STRING)
@click.pass_context
def permission_describe(ctx: click.Context, name: str):
    """
    Describe a permission
    """
    store = create_feature_store(ctx)

    try:
        permission = store.get_permission(name)
    except FeastObjectNotFoundException as e:
        print(e)
        exit(1)

    print(
        yaml.dump(
            yaml.safe_load(str(permission)), default_flow_style=False, sort_keys=False
        )
    )


@feast_permissions_cmd.command(name="check")
@click.pass_context
def feast_permissions_check_command(ctx: click.Context):
    """
    Validate the permissions configuration
    """
    from tabulate import tabulate

    all_unsecured_table: list[Any] = []
    store = create_feature_store(ctx)
    permissions = store.list_permissions()
    objects = cli_utils.fetch_all_feast_objects(
        store=store,
    )

    print(
        f"{Style.BRIGHT + Fore.RED}The following resources are not secured by any permission configuration:{Style.RESET_ALL}"
    )
    for o in objects:
        cli_utils.handle_permissions_check_command(
            object=o, permissions=permissions, table=all_unsecured_table
        )
    print(
        tabulate(
            all_unsecured_table,
            headers=[
                "NAME",
                "TYPE",
            ],
            tablefmt="plain",
        )
    )

    all_unsecured_actions_table: list[Any] = []
    print(
        f"{Style.BRIGHT + Fore.RED}The following actions are not secured by any permission configuration (Note: this might not be a security concern, depending on the used APIs):{Style.RESET_ALL}"
    )
    for o in objects:
        cli_utils.handle_permissions_check_command_with_actions(
            object=o, permissions=permissions, table=all_unsecured_actions_table
        )
    print(
        tabulate(
            all_unsecured_actions_table,
            headers=[
                "NAME",
                "TYPE",
                "UNSECURED ACTIONS",
            ],
            tablefmt="plain",
        )
    )


@feast_permissions_cmd.command(name="list-roles")
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Print the resources and actions permitted to each configured role",
)
@click.pass_context
def feast_permissions_list_roles_command(ctx: click.Context, verbose: bool):
    """
    List all the configured roles
    """
    from tabulate import tabulate

    table: list[Any] = []
    store = create_feature_store(ctx)
    permissions = store.list_permissions()
    if not verbose:
        cli_utils.handler_list_all_permissions_roles(
            permissions=permissions, table=table
        )
        print(
            tabulate(
                table,
                headers=[
                    "ROLE NAME",
                ],
                tablefmt="grid",
            )
        )
    else:
        objects = cli_utils.fetch_all_feast_objects(
            store=store,
        )
        cli_utils.handler_list_all_permissions_roles_verbose(
            objects=objects, permissions=permissions, table=table
        )
        print(
            tabulate(
                table,
                headers=[
                    "ROLE NAME",
                    "RESOURCE NAME",
                    "RESOURCE TYPE",
                    "PERMITTED ACTIONS",
                ],
                tablefmt="plain",
            )
        )



from datetime import datetime, timedelta
from typing import Callable, List, Optional, Union, cast

from pyspark.sql import DataFrame, SparkSession, Window
from pyspark.sql import functions as F

from feast import BatchFeatureView, StreamFeatureView
from feast.aggregation import Aggregation
from feast.data_source import DataSource
from feast.infra.common.serde import SerializedArtifacts
from feast.infra.compute_engines.dag.context import ColumnInfo, ExecutionContext
from feast.infra.compute_engines.dag.model import DAGFormat
from feast.infra.compute_engines.dag.node import DAGNode
from feast.infra.compute_engines.dag.value import DAGValue
from feast.infra.compute_engines.spark.utils import map_in_arrow
from feast.infra.compute_engines.utils import (
    create_offline_store_retrieval_job,
)
from feast.infra.offline_stores.contrib.spark_offline_store.spark import (
    SparkRetrievalJob,
    _get_entity_schema,
)
from feast.infra.offline_stores.contrib.spark_offline_store.spark_source import (
    SparkSource,
)
from feast.infra.offline_stores.offline_utils import (
    infer_event_timestamp_from_entity_df,
)

ENTITY_TS_ALIAS = "__entity_event_timestamp"


# Rename entity_df event_timestamp_col to match feature_df
def rename_entity_ts_column(
    spark_session: SparkSession, entity_df: DataFrame
) -> DataFrame:
    # check if entity_ts_alias already exists
    if ENTITY_TS_ALIAS in entity_df.columns:
        return entity_df

    entity_schema = _get_entity_schema(
        spark_session=spark_session,
        entity_df=entity_df,
    )
    event_timestamp_col = infer_event_timestamp_from_entity_df(
        entity_schema=entity_schema,
    )
    if not isinstance(entity_df, DataFrame):
        entity_df = spark_session.createDataFrame(entity_df)
    entity_df = entity_df.withColumnRenamed(event_timestamp_col, ENTITY_TS_ALIAS)
    return entity_df


class SparkReadNode(DAGNode):
    def __init__(
        self,
        name: str,
        source: DataSource,
        column_info: ColumnInfo,
        spark_session: SparkSession,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
    ):
        super().__init__(name)
        self.source = source
        self.column_info = column_info
        self.spark_session = spark_session
        self.start_time = start_time
        self.end_time = end_time

    def execute(self, context: ExecutionContext) -> DAGValue:
        retrieval_job = create_offline_store_retrieval_job(
            data_source=self.source,
            column_info=self.column_info,
            context=context,
            start_time=self.start_time,
            end_time=self.end_time,
        )
        if isinstance(retrieval_job, SparkRetrievalJob):
            spark_df = cast(SparkRetrievalJob, retrieval_job).to_spark_df()
        else:
            spark_df = self.spark_session.createDataFrame(retrieval_job.to_arrow())

        return DAGValue(
            data=spark_df,
            format=DAGFormat.SPARK,
            metadata={
                "source": "feature_view_batch_source",
                "timestamp_field": self.column_info.timestamp_column,
                "created_timestamp_column": self.column_info.created_timestamp_column,
                "start_date": self.start_time,
                "end_date": self.end_time,
            },
        )


class SparkAggregationNode(DAGNode):
    def __init__(
        self,
        name: str,
        aggregations: List[Aggregation],
        group_by_keys: List[str],
        timestamp_col: str,
        inputs=None,
    ):
        super().__init__(name, inputs=inputs)
        self.aggregations = aggregations
        self.group_by_keys = group_by_keys
        self.timestamp_col = timestamp_col

    def execute(self, context: ExecutionContext) -> DAGValue:
        input_value = self.get_single_input_value(context)
        input_value.assert_format(DAGFormat.SPARK)
        input_df: DataFrame = input_value.data

        agg_exprs = []
        for agg in self.aggregations:
            func = getattr(F, agg.function)
            expr = func(agg.column).alias(
                f"{agg.function}_{agg.column}_{int(agg.time_window.total_seconds())}s"
                if agg.time_window
                else f"{agg.function}_{agg.column}"
            )
            agg_exprs.append(expr)

        if any(agg.time_window for agg in self.aggregations):
            # 🕒 Use Spark's `window` function
            time_window = self.aggregations[
                0
            ].time_window  # assume consistent window size for now
            if time_window is None:
                raise ValueError("Aggregation requires time_window but got None.")
            window_duration_str = f"{int(time_window.total_seconds())} seconds"

            grouped = input_df.groupBy(
                *self.group_by_keys,
                F.window(F.col(self.timestamp_col), window_duration_str),
            ).agg(*agg_exprs)
        else:
            # Simple aggregation
            grouped = input_df.groupBy(
                *self.group_by_keys,
            ).agg(*agg_exprs)

        return DAGValue(
            data=grouped, format=DAGFormat.SPARK, metadata={"aggregated": True}
        )


class SparkJoinNode(DAGNode):
    def __init__(
        self,
        name: str,
        column_info: ColumnInfo,
        spark_session: SparkSession,
        inputs: Optional[List[DAGNode]] = None,
        how: str = "inner",
    ):
        super().__init__(name, inputs=inputs or [])
        self.column_info = column_info
        self.spark_session = spark_session
        self.how = how

    def execute(self, context: ExecutionContext) -> DAGValue:
        input_values = self.get_input_values(context)
        for val in input_values:
            val.assert_format(DAGFormat.SPARK)

        # Join all input DataFrames on join_keys
        joined_df = None
        for i, dag_value in enumerate(input_values):
            df = dag_value.data

            # Use original FeatureView name if available
            fv_name = self.inputs[i].name.split(":")[0]
            prefix = fv_name + "__"

            # Skip renaming join keys to preserve join compatibility
            renamed_cols = [
                F.col(c).alias(f"{prefix}{c}")
                if c not in self.column_info.join_keys
                else F.col(c)
                for c in df.columns
            ]
            df = df.select(*renamed_cols)
            if joined_df is None:
                joined_df = df
            else:
                joined_df = joined_df.join(
                    df, on=self.column_info.join_keys, how=self.how
                )

        # If entity_df is provided, join it in last
        entity_df = context.entity_df
        if entity_df is not None:
            entity_df = rename_entity_ts_column(
                spark_session=self.spark_session,
                entity_df=entity_df,
            )
            if joined_df is None:
                raise RuntimeError("No input features available to join with entity_df")

            joined_df = entity_df.join(
                joined_df, on=self.column_info.join_keys, how="left"
            )

        return DAGValue(
            data=joined_df,
            format=DAGFormat.SPARK,
            metadata={"joined_on": self.column_info.join_keys, "join_type": self.how},
        )


class SparkFilterNode(DAGNode):
    def __init__(
        self,
        name: str,
        column_info: ColumnInfo,
        spark_session: SparkSession,
        ttl: Optional[timedelta] = None,
        filter_condition: Optional[str] = None,
        inputs=None,
    ):
        super().__init__(name, inputs=inputs)
        self.column_info = column_info
        self.spark_session = spark_session
        self.ttl = ttl
        self.filter_condition = filter_condition

    def execute(self, context: ExecutionContext) -> DAGValue:
        input_value = self.get_single_input_value(context)
        input_value.assert_format(DAGFormat.SPARK)
        input_df: DataFrame = input_value.data

        # Get timestamp fields from feature view
        timestamp_column = self.column_info.timestamp_column

        # Optional filter: feature.ts <= entity.event_timestamp
        filtered_df = input_df
        if ENTITY_TS_ALIAS in input_df.columns:
            filtered_df = filtered_df.filter(
                F.col(timestamp_column) <= F.col(ENTITY_TS_ALIAS)
            )

            # Optional TTL filter: feature.ts >= entity.event_timestamp - ttl
            if self.ttl:
                ttl_seconds = int(self.ttl.total_seconds())
                lower_bound = F.col(ENTITY_TS_ALIAS) - F.expr(
                    f"INTERVAL {ttl_seconds} seconds"
                )
                filtered_df = filtered_df.filter(F.col(timestamp_column) >= lower_bound)

        # Optional custom filter condition
        if self.filter_condition:
            filtered_df = filtered_df.filter(self.filter_condition)

        return DAGValue(
            data=filtered_df,
            format=DAGFormat.SPARK,
            metadata={"filter_applied": True},
        )


class SparkDedupNode(DAGNode):
    def __init__(
        self,
        name: str,
        column_info: ColumnInfo,
        spark_session: SparkSession,
        inputs=None,
    ):
        super().__init__(name, inputs=inputs)
        self.column_info = column_info
        self.spark_session = spark_session

    def execute(self, context: ExecutionContext) -> DAGValue:
        input_value = self.get_single_input_value(context)
        input_value.assert_format(DAGFormat.SPARK)
        input_df: DataFrame = input_value.data

        # Dedup based on join keys and event timestamp column
        # Dedup with row_number
        partition_cols = self.column_info.join_keys
        deduped_df = input_df
        if partition_cols:
            ordering = [F.col(self.column_info.timestamp_column).desc()]
            if self.column_info.created_timestamp_column:
                ordering.append(F.col(self.column_info.created_timestamp_column).desc())

            window = Window.partitionBy(*partition_cols).orderBy(*ordering)
            deduped_df = (
                input_df.withColumn("row_num", F.row_number().over(window))
                .filter("row_num = 1")
                .drop("row_num")
            )

        return DAGValue(
            data=deduped_df,
            format=DAGFormat.SPARK,
            metadata={"deduped": True},
        )


class SparkWriteNode(DAGNode):
    def __init__(
        self,
        name: str,
        feature_view: Union[BatchFeatureView, StreamFeatureView],
        inputs=None,
    ):
        super().__init__(name, inputs=inputs)
        self.feature_view = feature_view

    def execute(self, context: ExecutionContext) -> DAGValue:
        spark_df: DataFrame = self.get_single_input_value(context).data
        serialized_artifacts = SerializedArtifacts.serialize(
            feature_view=self.feature_view, repo_config=context.repo_config
        )

        # ✅ 1. Write to online store if online enabled
        if self.feature_view.online:
            spark_df.mapInArrow(
                lambda x: map_in_arrow(x, serialized_artifacts, mode="online"),
                spark_df.schema,
            ).count()

        # ✅ 2. Write to offline store if offline enabled
        if self.feature_view.offline:
            if not isinstance(self.feature_view.batch_source, SparkSource):
                spark_df.mapInArrow(
                    lambda x: map_in_arrow(x, serialized_artifacts, mode="offline"),
                    spark_df.schema,
                ).count()
            # Directly write spark df to spark offline store without using mapInArrow
            else:
                dest_path = self.feature_view.batch_source.path
                file_format = self.feature_view.batch_source.file_format
                if not dest_path or not file_format:
                    raise ValueError(
                        "Destination path and file format must be specified for SparkSource."
                    )
                spark_df.write.format(file_format).mode("append").save(dest_path)

        return DAGValue(
            data=spark_df,
            format=DAGFormat.SPARK,
            metadata={
                "feature_view": self.feature_view.name,
                "write_to_online": self.feature_view.online,
                "write_to_offline": self.feature_view.offline,
            },
        )


class SparkTransformationNode(DAGNode):
    def __init__(self, name: str, udf: Callable, inputs: List[DAGNode]):
        super().__init__(name, inputs)
        self.udf = udf

    def execute(self, context: ExecutionContext) -> DAGValue:
        input_values = self.get_input_values(context)
        for val in input_values:
            val.assert_format(DAGFormat.SPARK)

        input_dfs: List[DataFrame] = [val.data for val in input_values]

        transformed_df = self.udf(*input_dfs)

        return DAGValue(
            data=transformed_df, format=DAGFormat.SPARK, metadata={"transformed": True}
        )



from fastapi import APIRouter, Depends, Query

from feast.api.registry.rest.rest_utils import (
    create_grpc_pagination_params,
    create_grpc_sorting_params,
    get_object_relationships,
    get_pagination_params,
    get_relationships_for_objects,
    get_sorting_params,
    grpc_call,
)
from feast.registry_server import RegistryServer_pb2


def get_permission_router(grpc_handler) -> APIRouter:
    router = APIRouter()

    @router.get("/permissions/{name}")
    def get_permission(
        name: str,
        project: str = Query(...),
        include_relationships: bool = Query(
            False, description="Include relationships for this permission"
        ),
        allow_cache: bool = Query(True),
    ):
        req = RegistryServer_pb2.GetPermissionRequest(
            name=name,
            project=project,
            allow_cache=allow_cache,
        )
        permission = grpc_call(grpc_handler.GetPermission, req)

        result = permission

        # Note: permissions may not have relationships in the traditional sense
        # but we include the functionality for consistency
        if include_relationships:
            relationships = get_object_relationships(
                grpc_handler, "permission", name, project, allow_cache
            )
            result["relationships"] = relationships

        return result

    @router.get("/permissions")
    def list_permissions(
        project: str = Query(...),
        allow_cache: bool = Query(default=True),
        include_relationships: bool = Query(
            False, description="Include relationships for each permission"
        ),
        pagination_params: dict = Depends(get_pagination_params),
        sorting_params: dict = Depends(get_sorting_params),
    ):
        req = RegistryServer_pb2.ListPermissionsRequest(
            project=project,
            allow_cache=allow_cache,
            pagination=create_grpc_pagination_params(pagination_params),
            sorting=create_grpc_sorting_params(sorting_params),
        )
        response = grpc_call(grpc_handler.ListPermissions, req)
        permissions = response.get("permissions", [])

        result = {
            "permissions": permissions,
            "pagination": response.get("pagination", {}),
        }

        if include_relationships:
            relationships = get_relationships_for_objects(
                grpc_handler, permissions, "permission", project, allow_cache
            )
            result["relationships"] = relationships

        return result

    return router



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
from datetime import datetime
from importlib.metadata import version as importlib_version
from pathlib import Path
from typing import List, Optional

import click
import yaml
from colorama import Fore, Style
from dateutil import parser
from pygments import formatters, highlight, lexers

from feast import utils
from feast.cli.data_sources import data_sources_cmd
from feast.cli.entities import entities_cmd
from feast.cli.feature_services import feature_services_cmd
from feast.cli.feature_views import feature_views_cmd
from feast.cli.features import (
    features_cmd,
    get_historical_features,
    get_online_features,
)
from feast.cli.on_demand_feature_views import on_demand_feature_views_cmd
from feast.cli.permissions import feast_permissions_cmd
from feast.cli.projects import projects_cmd
from feast.cli.saved_datasets import saved_datasets_cmd
from feast.cli.serve import (
    serve_command,
    serve_offline_command,
    serve_registry_command,
    serve_transformations_command,
)
from feast.cli.stream_feature_views import stream_feature_views_cmd
from feast.cli.ui import ui
from feast.cli.validation_references import validation_references_cmd
from feast.constants import FEAST_FS_YAML_FILE_PATH_ENV_NAME
from feast.errors import FeastProviderLoginError
from feast.repo_config import load_repo_config
from feast.repo_operations import (
    apply_total,
    cli_check_repo,
    create_feature_store,
    generate_project_name,
    init_repo,
    plan,
    registry_dump,
    teardown,
)
from feast.utils import maybe_local_tz

_logger = logging.getLogger(__name__)


class NoOptionDefaultFormat(click.Command):
    def format_options(self, ctx: click.Context, formatter: click.HelpFormatter):
        """Writes all the options into the formatter if they exist."""
        opts = []
        for param in self.get_params(ctx):
            rv = param.get_help_record(ctx)
            if rv is not None:
                opts.append(rv)
        if opts:
            with formatter.section("Options(No current command options)"):
                formatter.write_dl(opts)


@click.group()
@click.option(
    "--chdir",
    "-c",
    envvar="FEATURE_REPO_DIR_ENV_VAR",
    help="Switch to a different feature repository directory before executing the given subcommand. Can also be set via the FEATURE_REPO_DIR_ENV_VAR environment variable.",
)
@click.option(
    "--log-level",
    default="warning",
    help="The logging level. One of DEBUG, INFO, WARNING, ERROR, and CRITICAL (case-insensitive).",
)
@click.option(
    "--feature-store-yaml",
    "-f",
    help=f"Override the directory where the CLI should look for the feature_store.yaml file. Can also be set via the {FEAST_FS_YAML_FILE_PATH_ENV_NAME} environment variable.",
)
@click.pass_context
def cli(
    ctx: click.Context,
    chdir: Optional[str],
    log_level: str,
    feature_store_yaml: Optional[str],
):
    """
    Feast CLI

    For more information, see our public docs at https://docs.feast.dev/
    """
    ctx.ensure_object(dict)
    ctx.obj["CHDIR"] = Path.cwd() if chdir is None else Path(chdir).absolute()
    ctx.obj["FS_YAML_FILE"] = (
        Path(feature_store_yaml).absolute()
        if feature_store_yaml
        else utils.get_default_yaml_file_path(ctx.obj["CHDIR"])
    )
    try:
        level = getattr(logging, log_level.upper())
        logging.basicConfig(
            format="%(asctime)s %(name)s %(levelname)s: %(message)s",
            datefmt="%m/%d/%Y %I:%M:%S %p",
            level=level,
        )
        # Override the logging level for already created loggers (due to loggers being created at the import time)
        # Note, that format & datefmt does not need to be set, because by default child loggers don't override them

        # Also note, that mypy complains that logging.root doesn't have "manager" because of the way it's written.
        # So we have to put a type ignore hint for mypy.
        for logger_name in logging.root.manager.loggerDict:  # type: ignore
            if "feast" in logger_name:
                logger = logging.getLogger(logger_name)
                logger.setLevel(level)
    except Exception as e:
        raise e
    pass


@cli.command()
def version():
    """
    Display Feast SDK version
    """
    print(
        f'{Style.BRIGHT + Fore.BLUE}Feast SDK Version: {Style.BRIGHT + Fore.GREEN}"{importlib_version("feast")}"'
    )


@cli.command()
@click.argument("object_id")
@click.pass_context
def delete(ctx: click.Context, object_id: str):
    """
    Delete Feast Object
    """
    repo = ctx.obj["CHDIR"]
    fs_yaml_file = ctx.obj["FS_YAML_FILE"]
    cli_check_repo(repo, fs_yaml_file)
    store = create_feature_store(ctx)

    e = None
    object_type = None

    # Order matters if names can overlap between types,
    # though typically they shouldn't in a well-structured feature store.
    object_getters_and_types = [
        (store.get_entity, "Entity"),
        (store.get_feature_view, "FeatureView"),
        (store.get_feature_service, "FeatureService"),
        (store.get_data_source, "DataSource"),
        (store.get_saved_dataset, "SavedDataset"),
        (store.get_validation_reference, "ValidationReference"),
        (store.get_stream_feature_view, "StreamFeatureView"),
        (store.get_on_demand_feature_view, "OnDemandFeatureView"),
        # Add other get_* methods here if needed
    ]

    for getter, obj_type_str in object_getters_and_types:
        try:
            potential_e = getter(object_id)  # type: ignore[operator]
            if potential_e:
                e = potential_e
                object_type = obj_type_str
                break
        except Exception:
            pass

    if isinstance(e, list):
        e = e[0]
    if e:
        store.apply([e], objects_to_delete=[e], partial=False)
        print(
            f"{Style.BRIGHT + Fore.RED}Deleted {Style.BRIGHT + Fore.GREEN}{object_type} {Fore.YELLOW}{object_id} from {Fore.GREEN}{store.project}.{Style.RESET_ALL}"
        )
    else:
        print(
            f"{Style.BRIGHT + Fore.GREEN}Object not found. Deletion skipped.{Style.RESET_ALL}"
        )


@cli.command()
@click.pass_context
def configuration(ctx: click.Context):
    """
    Display Feast configuration
    """
    repo = ctx.obj["CHDIR"]
    fs_yaml_file = ctx.obj["FS_YAML_FILE"]
    cli_check_repo(repo, fs_yaml_file)
    repo_config = load_repo_config(repo, fs_yaml_file)
    if repo_config:
        config_dict = repo_config.model_dump(by_alias=True, exclude_unset=True)
        config_dict.pop("repo_path", None)
        print(yaml.dump(config_dict, default_flow_style=False, sort_keys=False))
    else:
        print("No configuration found.")


@cli.command()
@click.pass_context
def endpoint(ctx: click.Context):
    """
    Display feature server endpoints
    """
    store = create_feature_store(ctx)
    endpoint = store.get_feature_server_endpoint()
    if endpoint is not None:
        _logger.info(
            f"Feature server endpoint: {Style.BRIGHT + Fore.GREEN}{endpoint}{Style.RESET_ALL}"
        )
    else:
        _logger.info("There is no active feature server.")


@cli.command("plan", cls=NoOptionDefaultFormat)
@click.option(
    "--skip-source-validation",
    is_flag=True,
    help="Don't validate the data sources by checking for that the tables exist.",
)
@click.pass_context
def plan_command(ctx: click.Context, skip_source_validation: bool):
    """
    Create or update a feature store deployment
    """
    repo = ctx.obj["CHDIR"]
    fs_yaml_file = ctx.obj["FS_YAML_FILE"]
    cli_check_repo(repo, fs_yaml_file)
    repo_config = load_repo_config(repo, fs_yaml_file)
    try:
        plan(repo_config, repo, skip_source_validation)
    except FeastProviderLoginError as e:
        print(str(e))


@cli.command("apply", cls=NoOptionDefaultFormat)
@click.option(
    "--skip-source-validation",
    is_flag=True,
    help="Don't validate the data sources by checking for that the tables exist.",
)
@click.pass_context
def apply_total_command(ctx: click.Context, skip_source_validation: bool):
    """
    Create or update a feature store deployment
    """
    repo = ctx.obj["CHDIR"]
    fs_yaml_file = ctx.obj["FS_YAML_FILE"]
    cli_check_repo(repo, fs_yaml_file)

    repo_config = load_repo_config(repo, fs_yaml_file)
    try:
        apply_total(repo_config, repo, skip_source_validation)
    except FeastProviderLoginError as e:
        print(str(e))


@cli.command("teardown", cls=NoOptionDefaultFormat)
@click.pass_context
def teardown_command(ctx: click.Context):
    """
    Tear down deployed feature store infrastructure
    """
    repo = ctx.obj["CHDIR"]
    fs_yaml_file = ctx.obj["FS_YAML_FILE"]
    cli_check_repo(repo, fs_yaml_file)
    repo_config = load_repo_config(repo, fs_yaml_file)

    teardown(repo_config, repo)


@cli.command("registry-dump")
@click.pass_context
def registry_dump_command(ctx: click.Context):
    """
    Print contents of the metadata registry
    """
    repo = ctx.obj["CHDIR"]
    fs_yaml_file = ctx.obj["FS_YAML_FILE"]
    cli_check_repo(repo, fs_yaml_file)
    repo_config = load_repo_config(repo, fs_yaml_file)

    click.echo(registry_dump(repo_config, repo_path=repo))


@cli.command("materialize")
@click.argument("start_ts", required=False)
@click.argument("end_ts", required=False)
@click.option(
    "--views",
    "-v",
    help="Feature views to materialize",
    multiple=True,
)
@click.option(
    "--disable-event-timestamp",
    is_flag=True,
    help="Materialize all available data using current datetime as event timestamp (useful when source data lacks event timestamps)",
)
@click.pass_context
def materialize_command(
    ctx: click.Context,
    start_ts: Optional[str],
    end_ts: Optional[str],
    views: List[str],
    disable_event_timestamp: bool,
):
    """
    Run a (non-incremental) materialization job to ingest data into the online store. Feast
    will read all data between START_TS and END_TS from the offline store and write it to the
    online store. If you don't specify feature view names using --views, all registered Feature
    Views will be materialized.

    START_TS and END_TS should be in ISO 8601 format, e.g. '2021-07-16T19:20:01'

    If --disable-event-timestamp is used, timestamps are not required and all available data will be materialized using the current datetime as the event timestamp.
    """
    store = create_feature_store(ctx)

    if disable_event_timestamp:
        if start_ts or end_ts:
            raise click.UsageError(
                "Cannot specify START_TS or END_TS when --disable-event-timestamp is used"
            )
        now = datetime.now()
        # Query all available data and use current datetime as event timestamp
        start_date = datetime(
            1970, 1, 1
        )  # Beginning of time to capture all historical data
        end_date = now
    else:
        if not start_ts or not end_ts:
            raise click.UsageError(
                "START_TS and END_TS are required unless --disable-event-timestamp is used"
            )
        start_date = utils.make_tzaware(parser.parse(start_ts))
        end_date = utils.make_tzaware(parser.parse(end_ts))

    store.materialize(
        feature_views=None if not views else views,
        start_date=start_date,
        end_date=end_date,
        disable_event_timestamp=disable_event_timestamp,
    )


@cli.command("materialize-incremental")
@click.argument("end_ts")
@click.option(
    "--views",
    "-v",
    help="Feature views to incrementally materialize",
    multiple=True,
)
@click.pass_context
def materialize_incremental_command(ctx: click.Context, end_ts: str, views: List[str]):
    """
    Run an incremental materialization job to ingest new data into the online store. Feast will read
    all data from the previously ingested point to END_TS from the offline store and write it to the
    online store. If you don't specify feature view names using --views, all registered Feature
    Views will be incrementally materialized.

    END_TS should be in ISO 8601 format, e.g. '2021-07-16T19:20:01'
    """
    store = create_feature_store(ctx)
    store.materialize_incremental(
        feature_views=None if not views else views,
        end_date=utils.make_tzaware(datetime.fromisoformat(end_ts)),
    )


@cli.command("init")
@click.argument("PROJECT_DIRECTORY", required=False)
@click.option(
    "--minimal", "-m", is_flag=True, help="Create an empty project repository"
)
@click.option(
    "--template",
    "-t",
    type=click.Choice(
        [
            "local",
            "gcp",
            "aws",
            "snowflake",
            "spark",
            "postgres",
            "hbase",
            "cassandra",
            "hazelcast",
            "ikv",
            "couchbase",
            "milvus",
            "ray",
        ],
        case_sensitive=False,
    ),
    help="Specify a template for the created project",
    default="local",
)
def init_command(project_directory, minimal: bool, template: str):
    """Create a new Feast repository"""
    if not project_directory:
        project_directory = generate_project_name()

    if minimal:
        template = "minimal"

    init_repo(project_directory, template)


@cli.command("listen")
@click.option(
    "--address",
    "-a",
    type=click.STRING,
    default="localhost:50051",
    show_default=True,
    help="Address of the gRPC server",
)
@click.option(
    "--max_workers",
    "-w",
    type=click.INT,
    default=10,
    show_default=False,
    help="The maximum number of threads that can be used to execute the gRPC calls",
)
@click.option(
    "--registry_ttl_sec",
    "-r",
    help="Number of seconds after which the registry is refreshed",
    type=click.INT,
    default=5,
    show_default=True,
)
@click.pass_context
def listen_command(
    ctx: click.Context,
    address: str,
    max_workers: int,
    registry_ttl_sec: int,
):
    """Start a gRPC feature server to ingest streaming features on given address"""
    from feast.infra.contrib.grpc_server import get_grpc_server

    store = create_feature_store(ctx)
    server = get_grpc_server(address, store, max_workers, registry_ttl_sec)
    server.start()
    server.wait_for_termination()


@cli.command("validate")
@click.option(
    "--feature-service",
    "-f",
    help="Specify a feature service name",
)
@click.option(
    "--reference",
    "-r",
    help="Specify a validation reference name",
)
@click.option(
    "--no-profile-cache",
    is_flag=True,
    help="Do not store cached profile in registry",
)
@click.argument("start_ts")
@click.argument("end_ts")
@click.pass_context
def validate(
    ctx: click.Context,
    feature_service: str,
    reference: str,
    start_ts: str,
    end_ts: str,
    no_profile_cache,
):
    """
    Perform validation of logged features (produced by a given feature service) against provided reference.

    START_TS and END_TS should be in ISO 8601 format, e.g. '2021-07-16T19:20:01'
    """
    store = create_feature_store(ctx)

    _feature_service = store.get_feature_service(name=feature_service)
    _reference = store.get_validation_reference(reference)

    result = store.validate_logged_features(
        source=_feature_service,
        reference=_reference,
        start=maybe_local_tz(datetime.fromisoformat(start_ts)),
        end=maybe_local_tz(datetime.fromisoformat(end_ts)),
        throw_exception=False,
        cache_profile=not no_profile_cache,
    )

    if not result:
        print(f"{Style.BRIGHT + Fore.GREEN}Validation successful!{Style.RESET_ALL}")
        return

    errors = [e.to_dict() for e in result.report.errors]
    formatted_json = json.dumps(errors, indent=4)
    colorful_json = highlight(
        formatted_json, lexers.JsonLexer(), formatters.TerminalFormatter()
    )
    print(f"{Style.BRIGHT + Fore.RED}Validation failed!{Style.RESET_ALL}")
    print(colorful_json)
    exit(1)


cli.add_command(data_sources_cmd)
cli.add_command(entities_cmd)
cli.add_command(feature_services_cmd)
cli.add_command(feature_views_cmd)
cli.add_command(features_cmd)
cli.add_command(get_historical_features)
cli.add_command(get_online_features)
cli.add_command(on_demand_feature_views_cmd)
cli.add_command(feast_permissions_cmd)
cli.add_command(projects_cmd)
cli.add_command(saved_datasets_cmd)
cli.add_command(stream_feature_views_cmd)
cli.add_command(validation_references_cmd)
cli.add_command(ui)
cli.add_command(serve_command)
cli.add_command(serve_offline_command)
cli.add_command(serve_registry_command)
cli.add_command(serve_transformations_command)

if __name__ == "__main__":
    cli()



import logging
import os
from typing import Optional
from unittest.mock import Mock

import jwt
from fastapi import Request
from fastapi.security import OAuth2AuthorizationCodeBearer
from jwt import PyJWKClient
from starlette.authentication import (
    AuthenticationError,
)

from feast.permissions.auth.token_parser import TokenParser
from feast.permissions.auth_model import OidcAuthConfig
from feast.permissions.oidc_service import OIDCDiscoveryService
from feast.permissions.user import User

logger = logging.getLogger(__name__)


class OidcTokenParser(TokenParser):
    """
    A `TokenParser` to use an OIDC server to retrieve the user details.
    Server settings are retrieved from the `auth` configurationof the Feature store.
    """

    _auth_config: OidcAuthConfig

    def __init__(self, auth_config: OidcAuthConfig):
        self._auth_config = auth_config
        self.oidc_discovery_service = OIDCDiscoveryService(
            self._auth_config.auth_discovery_url
        )

    async def _validate_token(self, access_token: str):
        """
        Validate the token extracted from the header of the user request against the OAuth2 server.
        """
        # FastAPI's OAuth2AuthorizationCodeBearer requires a Request type but actually uses only the headers field
        # https://github.com/tiangolo/fastapi/blob/eca465f4c96acc5f6a22e92fd2211675ca8a20c8/fastapi/security/oauth2.py#L380
        request = Mock(spec=Request)
        request.headers = {"Authorization": f"Bearer {access_token}"}

        oauth_2_scheme = OAuth2AuthorizationCodeBearer(
            tokenUrl=self.oidc_discovery_service.get_token_url(),
            authorizationUrl=self.oidc_discovery_service.get_authorization_url(),
            refreshUrl=self.oidc_discovery_service.get_refresh_url(),
        )

        await oauth_2_scheme(request=request)

    async def user_details_from_access_token(self, access_token: str) -> User:
        """
        Validate the access token then decode it to extract the user credential and roles.

        Returns:
            User: Current user, with associated roles.

        Raises:
            AuthenticationError if any error happens.
        """

        # check if intra server communication
        user = self._get_intra_comm_user(access_token)
        if user:
            return user

        try:
            await self._validate_token(access_token)
            logger.debug("Token successfully validated.")
        except Exception as e:
            logger.error(f"Token validation failed: {e}")
            raise AuthenticationError(f"Invalid token: {e}")

        optional_custom_headers = {"User-agent": "custom-user-agent"}
        jwks_client = PyJWKClient(
            self.oidc_discovery_service.get_jwks_url(), headers=optional_custom_headers
        )

        try:
            signing_key = jwks_client.get_signing_key_from_jwt(access_token)
            data = jwt.decode(
                access_token,
                signing_key.key,
                algorithms=["RS256"],
                audience="account",
                options={
                    "verify_aud": False,
                    "verify_signature": True,
                    "verify_exp": True,
                },
                leeway=10,  # accepts tokens generated up to 10 seconds in the past, in case of clock skew
            )

            if "preferred_username" not in data:
                raise AuthenticationError(
                    "Missing preferred_username field in access token."
                )
            current_user = data["preferred_username"]

            if "resource_access" not in data:
                logger.warning("Missing resource_access field in access token.")
            client_id = self._auth_config.client_id
            if client_id not in data["resource_access"]:
                logger.warning(
                    f"Missing resource_access.{client_id} field in access token. Defaulting to empty roles."
                )
                roles = []
            else:
                roles = data["resource_access"][client_id]["roles"]

            logger.info(f"Extracted user {current_user} and roles {roles}")
            return User(username=current_user, roles=roles)
        except jwt.exceptions.InvalidTokenError:
            logger.exception("Exception while parsing the token:")
            raise AuthenticationError("Invalid token.")

    def _get_intra_comm_user(self, access_token: str) -> Optional[User]:
        intra_communication_base64 = os.getenv("INTRA_COMMUNICATION_BASE64")

        if intra_communication_base64:
            decoded_token = jwt.decode(
                access_token, options={"verify_signature": False}
            )
            if "preferred_username" in decoded_token:
                preferred_username: str = decoded_token["preferred_username"]
                if (
                    preferred_username is not None
                    and preferred_username == intra_communication_base64
                ):
                    return User(username=preferred_username, roles=[])

        return None



from feast.file_utils import replace_str_in_file


def bootstrap():
    import pathlib
    from datetime import datetime, timedelta

    import numpy as np
    import pandas as pd

    from feast.driver_test_data import create_driver_hourly_stats_df

    repo_path = pathlib.Path(__file__).parent.absolute() / "feature_repo"
    project_name = pathlib.Path(__file__).parent.absolute().name
    data_path = repo_path / "data"
    data_path.mkdir(exist_ok=True)

    end_date = datetime.now().replace(microsecond=0, second=0, minute=0)
    start_date = end_date - timedelta(days=15)

    # Generate driver data using Feast's built-in test data generator
    driver_entities = [1001, 1002, 1003, 1004, 1005]
    driver_df = create_driver_hourly_stats_df(driver_entities, start_date, end_date)

    if driver_df["event_timestamp"].dt.tz is None:
        driver_df["event_timestamp"] = driver_df["event_timestamp"].dt.tz_localize(
            "UTC"
        )
    if "created" in driver_df.columns and driver_df["created"].dt.tz is None:
        driver_df["created"] = driver_df["created"].dt.tz_localize("UTC")

    driver_stats_path = data_path / "driver_stats.parquet"
    driver_df.to_parquet(path=str(driver_stats_path), allow_truncated_timestamps=True)

    # Generate customer data to demonstrate Ray's multi-source capabilities
    customer_entities = [2001, 2002, 2003, 2004, 2005]

    # Create customer daily profile data
    customer_data = []
    for customer_id in customer_entities:
        for i, single_date in enumerate(
            pd.date_range(start_date, end_date, freq="D", tz="UTC")
        ):
            stable_timestamp = single_date.replace(
                hour=12, minute=0, second=0, microsecond=0
            )
            customer_data.append(
                {
                    "customer_id": customer_id,
                    "event_timestamp": stable_timestamp,
                    "created": stable_timestamp + timedelta(minutes=10),
                    "current_balance": np.random.uniform(10.0, 1000.0),
                    "avg_passenger_count": np.random.uniform(1.0, 4.0),
                    "lifetime_trip_count": np.random.randint(50, 500),
                }
            )

    customer_df = pd.DataFrame(customer_data)

    if customer_df["event_timestamp"].dt.tz is None:
        customer_df["event_timestamp"] = customer_df["event_timestamp"].dt.tz_localize(
            "UTC"
        )
    if customer_df["created"].dt.tz is None:
        customer_df["created"] = customer_df["created"].dt.tz_localize("UTC")

    customer_profile_path = data_path / "customer_daily_profile.parquet"
    customer_df.to_parquet(
        path=str(customer_profile_path), allow_truncated_timestamps=True
    )

    # Update the example_repo.py file with actual paths
    example_py_file = repo_path / "example_repo.py"
    replace_str_in_file(example_py_file, "%PROJECT_NAME%", str(project_name))
    replace_str_in_file(
        example_py_file, "%PARQUET_PATH%", str(driver_stats_path.relative_to(repo_path))
    )
    replace_str_in_file(
        example_py_file, "%LOGGING_PATH%", str(data_path.relative_to(repo_path))
    )

    print("Ray template initialized with sample data:")
    print(f"  - Driver stats: {driver_stats_path}")
    print(f"  - Customer profiles: {customer_profile_path}")
    print(f"  - Ray storage will be created at: {data_path / 'ray_storage'}")
    print("\nTo get started:")
    print(f"  1. cd {project_name}/feature_repo")
    print("  2. feast apply")
    print("  3. python test_workflow.py")


if __name__ == "__main__":
    bootstrap()



from abc import ABC, abstractmethod
from datetime import timedelta


class DataFrameBackend(ABC):
    """
    Abstract interface for DataFrame operations used by the LocalComputeEngine.

    This interface defines the contract for implementing pluggable DataFrame backends
    such as Pandas, Polars, or DuckDB. Each backend must support core table operations
    such as joins, filtering, aggregation, conversion to/from Arrow, and deduplication.

    The purpose of this abstraction is to allow seamless swapping of execution backends
    without changing DAGNode or ComputeEngine logic. All nodes operate on pyarrow.Table
    as the standard input/output format, while the backend defines how the computation
    is actually performed.

    Expected implementations include:
    - PandasBackend
    - PolarsBackend
    - DuckDBBackend (future)

    Methods
    -------
    from_arrow(table: pa.Table) -> Any
        Convert a pyarrow.Table to the backend-native DataFrame format.

    to_arrow(df: Any) -> pa.Table
        Convert a backend-native DataFrame to pyarrow.Table.

    join(left: Any, right: Any, on: List[str], how: str) -> Any
        Join two dataframes on specified keys with given join type.

    groupby_agg(df: Any, group_keys: List[str], agg_ops: Dict[str, Tuple[str, str]]) -> Any
        Group and aggregate the dataframe. `agg_ops` maps output column names
        to (aggregation function, source column name) pairs.

    filter(df: Any, expr: str) -> Any
        Apply a filter expression (string-based) to the DataFrame.

    to_timedelta_value(delta: timedelta) -> Any
        Convert a Python timedelta object to a backend-compatible value
        that can be subtracted from a timestamp column.

    drop_duplicates(df: Any, keys: List[str], sort_by: List[str], ascending: bool = False) -> Any
        Deduplicate the DataFrame by key columns, keeping the first row
        by descending or ascending sort order.

    rename_columns(df: Any, columns: Dict[str, str]) -> Any
        Rename columns in the DataFrame according to the provided mapping.
    """

    @abstractmethod
    def columns(self, df): ...

    @abstractmethod
    def from_arrow(self, table): ...

    @abstractmethod
    def join(self, left, right, on, how): ...

    @abstractmethod
    def groupby_agg(self, df, group_keys, agg_ops): ...

    @abstractmethod
    def filter(self, df, expr): ...

    @abstractmethod
    def to_arrow(self, df): ...

    @abstractmethod
    def to_timedelta_value(self, delta: timedelta): ...

    @abstractmethod
    def drop_duplicates(self, df, keys, sort_by, ascending: bool = False):
        pass

    @abstractmethod
    def rename_columns(self, df, columns: dict[str, str]): ...
